#
#    Copyright (C) 2010-2016 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
TODO: documentation here.
"""
from pytrip.tripexecuter.pytripobj import pytripObj


class TripPlanCollection(pytripObj):
    def __init__(self):
        self.save_fields = ["plans"]
        self.plans = []
        self.active_plan = None

    def get_plans(self):
        return self.plans

    def add_plan(self, plan):
        plan.Init(self)
        if plan.name == "":
            plan.set_name("plan %d" % (len(self.plans) + 1))
        self.plans.append(plan)
        self.active_plan = plan

    def __iter__(self):
        self.iter = 0
        return self

    def next(self):
        if self.iter >= len(self.plans):
            raise StopIteration
        self.iter += 1
        return self.plans[self.iter - 1]

    def get_plan_by_id(self, id):
        return self.plans[id]

    def get_plan(self, name):
        for plan in self.plans:
            if plan.name == name:
                return plan
        return None

    def get_active_plan(self):
        return self.active_plan

    def set_active_plan(self, plan):
        self.active_plan = plan

    def delete_plan(self, plan):
        self.plans.remove(plan)

        if self.active_plan is plan:
            new_plan = None
            if len(self.plans) is not 0:
                new_plan = self.plans[0]
            return new_plan

        return None

    def delete_plan_by_id(self, id):
        plan = self.plans.pop(id)
        plan.destroy()



#
#    Copyright (C) 2010-2016 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This module provides the DosCube class, which the user should use for handling plan-generated dose distributions.
"""
import os
import logging

import numpy as np

from pytrip.error import InputError, ModuleNotLoadedError
from pytrip.cube import Cube
import pytriplib

try:
    from dicom.dataset import Dataset, FileDataset
    from dicom.sequence import Sequence
    from dicom import UID

    _dicom_loaded = True
except:
    _dicom_loaded = False

logger = logging.getLogger(__name__)


class DosCube(Cube):
    """ Class for handling Dose data. In TRiP98 these are stored in VOXELPLAN format with the .dos suffix.
    This class can also handle Dicom files.
    """
    data_file_extension = ".dos"
    header_file_extension = ".hed"

    def __init__(self, cube=None):
        super(DosCube, self).__init__(cube)
        self.type = "DOS"
        self.target_dose = 0.0

    def read_dicom(self, dcm):
        """ Imports the dose distribution from Dicom object.

        :param Dicom dcm: a Dicom object
        """
        if "rtdose" not in dcm:
            raise InputError("Data doesn't contain dose infomation")
        if self.header_set is False:
            self.read_dicom_header(dcm)
        self.cube = np.zeros((self.dimz, self.dimy, self.dimx))
        for i, item in enumerate(dcm["rtdose"].pixel_array):
            self.cube[i][:][:] = item

    def calculate_dvh(self, voi):
        """
        Calculate DHV for given VOI. Dose is given in relative units (target dose = 1.0).
        In case VOI lies outside the cube, then None is returned.

        :param voi: VOI for which DHV should be calculated
        :return: (dvh, min_dose, max_dose, mean, area) tuple. dvh - 2D array holding DHV histogram,
        min_dose and max_dose, mean_dose - obvious, mean_volume - effective volume dose.
        """

        z_pos = 0  # z position
        voxel_size = np.array([self.pixel_size, self.pixel_size, self.slice_distance])
        # in TRiP98 dose is stored in relative numbers, target dose is set to 1000 (and stored as 2-bytes ints)
        maximum_dose = 1500  # do not change, same value is hardcoded in filter_point.c (calculate_dvh_slice method)
        dose_bins = np.zeros(maximum_dose)  # placeholder for DVH, filled with zeros
        voi_and_cube_intersect = False
        for i in range(self.dimz):
            z_pos += self.slice_distance
            slice = voi.get_slice_at_pos(z_pos)
            if slice is not None:   # VOI intersects with this slice
                voi_and_cube_intersect = True
                dose_bins += pytriplib.calculate_dvh_slice(self.cube[i], np.array(slice.contour[0].contour), voxel_size)

        if voi_and_cube_intersect:
            sum_of_doses = sum(dose_bins)
            # np.cumsum - cumulative sum of array along the axis
            # we calculate is backwards and revert to get a plot which is monotonically decreasing
            # normalization is needed to get maximum values on Y axis to be <= 1
            dvh_x = np.arange(start=0.0, stop=1500.0)
            dvh_y = np.cumsum(dose_bins[::-1])[::-1] / sum_of_doses

            min_dose = np.where(dvh_y >= 0.98)[0][-1]
            max_dose = np.where(dvh_y <= 0.02)[0][0]

            # mean = \sum_i=1^1500 f_i * d_i , where:
            #   f_i = dose_bin(i) / \sum_i dose_bin(i)  - frequency of dose at index i
            #   d_i = dvx_i(i) = i                      - dose at index i (equal to i)
            #             dose goes from 0 to 1500 and is integer
            mean_dose = np.dot(dose_bins, dvh_x) / sum_of_doses

            # if full voi is irradiated with target dose, then it should be equal to VOI volume
            mean_volume = sum_of_doses * voxel_size[0] * voxel_size[1] * voxel_size[2]

            # TRiP98 target dose is 1000, we renormalize to 1.0
            min_dose /= 1000.0
            max_dose /= 1000.0
            mean_dose /= 1000.0
            mean_volume /= 1000.0
            dvh_x /= 1000.0

            dvh = np.column_stack((dvh_x, dvh_y))

            return dvh, min_dose, max_dose, mean_dose, mean_volume
        return None

    def write_dvh(self, voi, filename):
        """
        Save DHV for given VOI to the file.
        """
        dvh_tuple = self.calculate_dvh(voi)
        if dvh_tuple is None:
            logger.warning("Voi {:s} outside the cube".format(voi.get_name()))
        else:
            dvh, min_dose, max_dose, mean_dose, mean_volume = dvh_tuple
            np.savetxt(dvh, filename)

    def create_dicom_plan(self):
        """ Create a dummy Dicom RT-plan object.

        The only data which is forwarded to this object, is self.patient_name.
        :returns: a Dicom RT-plan object.
        """
        meta = Dataset()
        meta.MediaStorageSOPClassUID = '1.2.840.10008.5.1.4.1.1.2'  # CT Image Storage
        meta.MediaStorageSOPInstanceUID = "1.2.3"
        meta.ImplementationClassUID = "1.2.3.4"
        meta.TransferSyntaxUID = UID.ImplicitVRLittleEndian  # Implicit VR Little Endian - Default Transfer Syntax
        ds = FileDataset("file", {}, file_meta=meta, preamble=b"\0" * 128)
        ds.PatientsName = self.patient_name
        ds.PatientID = "123456"
        ds.PatientsSex = '0'
        ds.PatientsBirthDate = '19010101'
        ds.SpecificCharacterSet = 'ISO_IR 100'
        ds.SOPClassUID = '1.2.840.10008.5.1.4.1.1.2'  # CT Image Storage
        ds.StudyInstanceUID = '1.2.3'
        ds.SOPInstanceUID = '1.2.3'

        ds.Modality = "RTPLAN"
        ds.SeriesDescription = 'RT Plan'
        ds.SeriesInstanceUID = '2.16.840.1.113662.2.12.0.3057.1241703565.43'
        ds.RTPlanDate = '19010101'
        ds.RTPlanGeometry = ''
        ds.RTPlanLabel = 'B1'
        ds.RTPlanTime = '000000'
        structure_ref = Dataset()
        structure_ref.RefdSOPClassUID = '1.2.840.10008.5.1.4.1.1.481.3'  # RT Structure Set Storage
        structure_ref.RefdSOPInstanceUID = '1.2.3'
        ds.RefdStructureSets = Sequence([structure_ref])

        dose_ref = Dataset()
        dose_ref.DoseReferenceNumber = 1
        dose_ref.DoseReferenceStructureType = 'SITE'
        dose_ref.DoseReferenceType = 'TARGET'
        dose_ref.TargetPrescriptionDose = self.target_dose
        dose_ref.DoseReferenceDescription = "TUMOR"
        ds.DoseReferences = Sequence([dose_ref])
        return ds

    def create_dicom(self):
        """ Creates a Dicom RT-Dose object from self.

        This function can be used to convert a TRiP98 Dose file to Dicom format.

        :returns: a Dicom RT-Dose object.
        """

        if not _dicom_loaded:
            raise ModuleNotLoadedError("Dicom")
        if not self.header_set:
            raise InputError("Header not loaded")

        ds = self.create_dicom_base()
        ds.Modality = 'RTDOSE'
        ds.SamplesperPixel = 1
        ds.BitsAllocated = self.num_bytes * 8
        ds.BitsStored = self.num_bytes * 8
        ds.AccessionNumber = ''
        ds.SeriesDescription = 'RT Dose'
        ds.DoseUnits = 'GY'
        ds.DoseType = 'PHYSICAL'
        ds.DoseGridScaling = self.target_dose / 10**5
        ds.DoseSummationType = 'PLAN'
        ds.SliceThickness = ''
        ds.InstanceCreationDate = '19010101'
        ds.InstanceCreationTime = '000000'
        ds.NumberOfFrames = len(self.cube)
        ds.PixelRepresentation = 0
        ds.StudyID = '1'
        ds.SeriesNumber = 14
        ds.GridFrameOffsetVector = [x * self.slice_distance for x in range(self.dimz)]
        ds.InstanceNumber = ''
        ds.NumberofFrames = len(self.cube)
        ds.PositionReferenceIndicator = "RF"
        ds.TissueHeterogeneityCorrection = ['IMAGE', 'ROI_OVERRIDE']
        ds.ImagePositionPatient = ["%.3f" % (self.xoffset * self.pixel_size), "%.3f" % (self.yoffset * self.pixel_size),
                                   "%.3f" % (self.slice_pos[0])]
        ds.SOPClassUID = '1.2.840.10008.5.1.4.1.1.481.2'
        ds.SOPInstanceUID = '1.2.246.352.71.7.320687012.47206.20090603085223'
        ds.SeriesInstanceUID = '1.2.246.352.71.2.320687012.28240.20090603082420'

        # Bind to rtplan
        rt_set = Dataset()
        rt_set.RefdSOPClassUID = '1.2.840.10008.5.1.4.1.1.481.5'
        rt_set.RefdSOPInstanceUID = '1.2.3'
        ds.ReferencedRTPlans = Sequence([rt_set])
        pixel_array = np.zeros((len(self.cube), ds.Rows, ds.Columns), dtype=self.pydata_type)
        pixel_array[:][:][:] = self.cube[:][:][:]
        ds.PixelData = pixel_array.tostring()
        return ds

    def write(self, path):
        """
        Write Dose data to disk, in TRiP98/Voxelplan format.

        This method will build and write both the .hed and .dos file.

        :param str path: Path, any file extentions will be ignored.
        """

        header_file, dos_file = self.parse_path(path_name=path)
        self.write_trip_header(header_file)
        self.write_trip_data(dos_file)

    def write_dicom(self, directory):
        """ Write Dose-data to disk, in Dicom format.

        This file will save the dose cube and a plan associated with that dose.
        Function call create_dicom() and create_dicom_plan() and then save these.

        :param str directory: Directory where 'rtdose.dcm' and 'trplan.dcm' will be stored.
        """
        dcm = self.create_dicom()
        plan = self.create_dicom_plan()
        dcm.save_as(os.path.join(directory, "rtdose.dcm"))
        plan.save_as(os.path.join(directory, "rtplan.dcm"))



#
#    Copyright (C) 2010-2016 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This module provides the DDD class for handling depth-dose curve kernels for TRiP98.
"""
import glob
import numpy as np
from scipy import interpolate


class DDD:
    """ Class for handling Depth-Dose Data.
    """
    def __init__(self):
        pass

    def get_ddd_data(self, energy, points):
        """ TODO: documentation
        """
        e = self.get_nearest_energy(energy)
        return interpolate.splev(points, self.ddd_data[e])

    def get_dist(self, energy):
        """ TODO: documentation
        """
        return interpolate.splev(energy, self.max_dist)

    def get_ddd_by_energy(self, energy, points):
        """ TODO: documentation
        """
        ev_point = np.array([points, [energy] * len(points)])
        return interpolate.griddata(self.points, self.ddd_list, np.transpose(ev_point), method='linear')

    def get_ddd_grid(self, energy_list, n):
        """ TODO: documentation
        """
        # ddd_list = []
        energy = []
        dist = []
        # point = []
        data = []

        ddd_e = self.ddd.keys()
        ddd_e = sorted(ddd_e)

        for e in energy_list:
            idx = np.where((np.array(ddd_e) >= e))[0][0] - 1

            d_lower = self.ddd[ddd_e[idx]]
            d_upper = self.ddd[ddd_e[idx + 1]]

            lower_idx = np.where(max(d_lower[1, :]) == d_lower[1, :])[0][0]
            upper_idx = np.where(max(d_upper[1, :]) == d_upper[1, :])[0][0]

            offset = 1 / (ddd_e[idx + 1] - ddd_e[idx]) * (e - ddd_e[idx + 1])
            x_offset = (d_upper[0, upper_idx] - d_lower[0, lower_idx]) * offset
            y_offset = 1 + (1 - d_upper[1, upper_idx] / d_lower[1, lower_idx]) * offset

            depth = d_upper[0, :] + x_offset
            ddd = d_upper[1, :] * y_offset
            xi = np.linspace(0, depth[-1], n)
            spl = interpolate.splrep(depth, ddd)
            data.extend(interpolate.splev(xi, spl))
            dist.extend(xi)
            energy.extend([e] * n)

        out = [dist, energy, data]
        return np.reshape(np.transpose(out), (len(energy_list), n, 3))

        # TODO why it is not used ?
        # ddd_list = []
        # energy = []
        # dist = []
        # point = []
        # for e in energy_list:
        #     dist.extend(np.linspace(0, self.get_dist(e), n))
        #     energy.extend([e] * n)
        # point.append(dist)
        # point.append(energy)
        # data = interpolate.griddata(self.points,
        #                             self.ddd_list,
        #                             np.transpose(point),
        #                             method='linear')
        # out = [dist, energy, data]
        # return np.reshape(np.transpose(out), (len(energy_list), n, 3))

    def load_ddd(self, directory):
        """ Loads all .ddd files found in 'directory'

        :params str directory: directory where the .ddd files are found.
        """
        x_data = []
        y_data = []
        points = [[], []]
        ddd = {}
        max_dist = []
        items = glob.glob(directory)
        for item in items:
            x_data = []
            y_data = []
            with open(item, 'r') as f:
                data = f.read()
            lines = data.split('\n')
            for n, line in enumerate(data.split("\n")):
                if line.find("energy") is not -1:
                    energy = float(line.split()[1])
                if line.find('!') is -1 and line.find('#') is -1:
                    break
            for i in range(n, len(lines)):
                if len(lines[i]) < 3:
                    continue
                point = [float(s) for s in lines[i].split()]
                x_data.append(point[0] * 10)
                y_data.append(point[1])
            ddd[energy] = np.array([x_data, y_data])
            max_dist.append([energy, x_data[-1]])

        max_dist = np.array(sorted(max_dist, key=lambda x: x[0]))
        self.max_dist = interpolate.splrep(max_dist[:, 0], max_dist[:, 1], s=0)
        #        spl_points = 1000
        #        xi = np.linspace(0, max_dist[-1, 1], spl_points)
        ddd_list = []
        for key, value in ddd.iteritems():
            points[0].extend(value[0])
            points[1].extend(len(value[0]) * [key])
            ddd_list.extend(value[1])
        self.ddd_list = ddd_list
        self.ddd = ddd
        self.points = np.transpose(points)



#! /usr/bin/env python
#
#    Copyright (C) 2010-2016 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Converts GD files to xmgrace files in .agr format.
"""
import sys
import argparse
import logging

import pytrip as pt
from pytrip.utils.gd2dat import ReadGd


def main(args=sys.argv[1:]):
    """ Main function for gd2agr.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("gd_file", help="location of gd file", type=str)
    parser.add_argument("dat_file", help="location of .dat to write", type=str, nargs='?')
    parser.add_argument("-v", "--verbosity", action='count', help="increase output verbosity", default=0)
    parser.add_argument('-V', '--version', action='version', version=pt.__version__)
    args = parser.parse_args(args)

    gd_data = ReadGd(args.gd_file, let=False, exp=True, agr=True)
    gd_data.export(args.dat_file)

    return 0


if __name__ == '__main__':
    logging.basicConfig()
    sys.exit(main(sys.argv[1:]))



#
#    Copyright (C) 2010-2016 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
TODO: documentation here.
"""
import numpy as np
from math import pi, sin, cos, acos, asin


def get_basis_from_angles(gantry, couch):
    gantry /= 180.0 / pi
    couch /= -180.0 / pi
    a = -np.array([sin(gantry) * cos(couch), -cos(gantry), sin(couch) * sin(gantry)])
    c = -np.array([sin(gantry + pi / 2) * cos(couch), -cos(gantry + pi / 2), sin(couch) * sin(gantry + pi / 2)])
    b = np.cross(a, c)
    return [a, b, c]


def angles_from_trip(gantry, couch):
    gantry += 90
    couch = -(couch + 90)
    return gantry, couch


def angles_to_trip(gantry, couch):
    gantry -= 90
    couch = -couch - 90
    return gantry, couch


def vector_to_angles(vec):
    gantry = acos(vec[1])
    couch = asin(vec[2] / sin(gantry))
    return gantry / pi * 180, couch / pi * 180


def point_in_polygon(x, y, polygon):
    intersects = 0
    n = len(polygon)
    x1 = polygon[0][0]
    y1 = polygon[0][1]
    for i in range(n + 1):
        x2 = polygon[i % n][0]
        y2 = polygon[i % n][1]
        if y > min(y1, y2):
            if y <= max(y1, y2):
                if x <= max(x1, x2):
                    if y1 != y2:
                        xinters = (y - y1) * (x2 - x1) / (y2 - y1) + x1
                    if x1 == x2 or x <= xinters:
                        intersects += 1
        x1 = x2
        y1 = y2
    if intersects % 2 != 0:
        return True
    else:
        return False


def get_nearest_point(point, contour):
    length = 100000000.0
    p_out = None
    p2 = np.array(point)
    for p in contour:
        temp_len = sum((p2 - np.array(p))**2)
        if temp_len < length:
            length = temp_len
            p_out = p
    return p_out


def max_list(a, b):
    return [max(x, y) for x, y in zip(a, b)]


def min_list(a, b):
    return [min(x, y) for x, y in zip(a, b)]


def array_to_point_array(points, offset):
    point = [[points[3 * i] - offset[0], points[3 * i + 1] - offset[1], points[3 * i + 2] - offset[2]]
             for i in range(int(len(points) / 3))]
    return point


def get_area_contour(polygon):
    points = np.zeros((len(polygon), 2))
    points[0:len(polygon)] = np.array(polygon)[:, 0:2]
    points[-1] = np.array(polygon[0])
    dx_dy = np.array([points[i + 1] - points[i] for i in range(len(points) - 1)])
    points = np.array([(points[i + 1] + points[i]) / 2 for i in range(len(points) - 1)])
    area = -sum(points[:, 1] * dx_dy[:, 0])
    return area


def get_x_intersection(y, polygon):
    intersections = []
    x1 = polygon[0][0]
    y1 = polygon[0][1]
    n = len(polygon)
    for i in range(n + 1):
        x2 = polygon[i % n][0]
        y2 = polygon[i % n][1]
        if y > min(y1, y2):
            if y <= max(y1, y2):
                x = (x2 - x1) / (y2 - y1) * (y - y1) + x1
                intersections.append(x)
        x1 = x2
        y1 = y2
    return intersections


# find a short distance probably not the shortest
def short_distance_polygon_idx(poly1, poly2):
    d = 10000000
    n1 = len(poly1)
    n2 = len(poly2)
    i1 = 0
    i2 = 0
    for i in range(n2):
        d1 = (poly2[i][0] - poly1[i1][0])**2 + (poly2[i][1] - poly1[i1][1])**2
        if d1 < d:
            i2 = i
            d = d1
    for i in range(n1):
        d2 = (poly2[i2][0] - poly1[i][0])**2 + (poly2[i2][1] - poly1[i][1])**2
        if d2 < d:
            i1 = i
            d = d2
    for i in range(n2):
        d1 = (poly2[i][0] - poly1[i1][0])**2 + (poly2[i][1] - poly1[i1][1])**2
        if d1 < d:
            i2 = i
            d = d1
    for i in range(n1):
        d2 = (poly2[i2][0] - poly1[i][0])**2 + (poly2[i2][1] - poly1[i][1])**2
        if d2 < d:
            i1 = i
            d = d2

    return i1, i2, float(d)**0.5



#
#    Copyright (C) 2010-2016 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Generates .png files for each slice found in the given cube.
"""
import os
import sys
import argparse
import logging

from numpy import arange, ma

import pytrip as pt

logger = logging.getLogger(__name__)


def load_data_cube(filename):
    """ Loads either a Dos or LET-cube.

    :params str filname: path to Dos or LET-cube.
    :returns: a DosCube or LETCube object and a str containing the path to the basename.
    """

    if not filename:
        logger.warn("Empty data cube filename")
        return None, None

    logger.info("Reading " + filename)

    d = None
    basename_cube = None

    # try to parse filename, first for LET then DOS cube
    # LET cube file have extension .dosemlet.dos, DOS cube files have extension .dos
    # this means that LET cube file can be interpreted as DOS cube file
    # it is then important to check first for LET cube
    #
    # looping over classes is used in case support for other cubes needs to be added and to avoid code duplication
    for cube_cls in (pt.LETCube, pt.DosCube):
        data_header, _ = cube_cls.parse_path(filename)  # parse input file path
        logger.debug("Checking if {:s} fits with {:s} class".format(filename, cube_cls.__name__))
        if data_header is not None:  # parsing successful
            logger.debug("Extracted data header {:s}".format(data_header))

            # extracting basename
            # we are not using python splitext function, as LET cube has double extension: .dosemlet.dos
            basename_end_index = data_header.rfind(cube_cls.header_file_extension)
            basename_cube = data_header[:basename_end_index]

            d = cube_cls()  # creating cube object, reading file will be done later
            break  # escaping from the loop

    if d is not None:  # parsing successful, we can proceed to reading the file
        d.read(filename)
        logger.info("Data cube shape" + str(d.cube.shape))
        if isinstance(d, pt.DosCube):
            d *= 0.1  # convert %% to %

        dmax = d.cube.max()
        dmin = d.cube.min()
        logger.info("Data min, max values: {:g} {:g}".format(dmin, dmax))
    else:
        logger.warn("Filename " + filename + " is neither valid DOS nor LET cube")

    return d, basename_cube


def load_ct_cube(filename):
    """ loads the CT cube

    :params str filename: path to filename which must be loaded
    :returns: a CtxCube object and a str containing the path to the basename.
    """

    if not filename:
        logger.warn("Empty CT cube filename")
        return None, None

    logger.info("Reading " + filename)
    ctx_header, _ = pt.CtxCube.parse_path(filename)

    if ctx_header is None:
        logger.warn("Path " + filename + " doesn't seem to point to proper CT cube")
        return None, None

    basename_cube = os.path.splitext(ctx_header)[0]
    c = pt.CtxCube()
    c.read(filename)
    logger.info("CT cube shape" + str(c.cube.shape))

    cmax = c.cube.max()
    cmin = c.cube.min()
    logger.info("CT min, max values: {:d} {:d}".format(cmin, cmax))

    return c, basename_cube


def main(args=sys.argv[1:]):
    """ The main function for cubeslice.py
    """

    # there are some cases when this script is run on systems without DISPLAY variable being set
    # in such case matplotlib backend has to be explicitly specified
    # we do it here and not in the top of the file, as interleaving imports with code lines is discouraged
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    from matplotlib import colors
    from matplotlib.ticker import MultipleLocator

    # parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("--data", help="data cube(dos, let etc)", type=str, nargs='?')
    parser.add_argument("--ct", help="CT cube", type=str, nargs='?')
    parser.add_argument(
        "-f", "--from", type=int, dest='sstart', metavar='N', help="Output from slice number N", default=1)
    parser.add_argument("-t", "--to", type=int, dest='sstop', metavar='M', help="Output up to slice number M")
    parser.add_argument("-H", "--HUbar", dest='HUbar', default=False, action='store_true', help="Add HU colour bar")
    parser.add_argument("-o", "--outputdir", dest='outputdir',
                        help="Write resulting files to this directory.", type=str, default=None)
    parser.add_argument('-v', '--verbosity', action='count', help="increase output verbosity", default=0)
    parser.add_argument('-V', '--version', action='version', version=pt.__version__)
    args = parser.parse_args(args)

    if args.verbosity == 1:
        logger.setLevel('INFO')
    if args.verbosity > 1:
        logger.setLevel('DEBUG')

    if args.data is None and args.ct is None:
        logger.error("Provide location of at least one cube")
        return 2

    # Check if different output path was requested. If yes, then check if it exists.
    if args.outputdir is not None:
        if os.path.isdir(args.outputdir) is False:
            logger.error("Output directory " + args.outputdir + "does not exist.")
            return 1

    data_cube, data_basename = load_data_cube(args.data)

    ct_cube, ct_basename = load_ct_cube(args.ct)

    # check cube data are compatible (if both cubes present)
    if data_cube is not None and ct_cube is not None:
        if not data_cube.is_compatible(ct_cube):
            logger.error("Cubes don't match")
            logger.info("Cube 1 " + args.data + " shape {:d} x {:d} x {:d}".format(data_cube.dimx, data_cube.dimy,
                                                                                   data_cube.dimz))
            logger.info("Cube 1 " + args.data + " pixel {:g} [cm], slice {:g} [cm]".format(data_cube.pixel_size,
                                                                                           data_cube.slice_distance))
            logger.info("Cube 2 " + args.ct + " shape {:d} x {:d} x {:d}".format(ct_cube.dimx, ct_cube.dimy,
                                                                                 ct_cube.dimz))
            logger.info("Cube 2 " + args.ct + " pixel {:g} [cm], slice {:g} [cm]".format(ct_cube.pixel_size,
                                                                                         ct_cube.slice_distance))
            return 2

    cube = None
    cube_basename = None
    if data_cube is not None:  # user provided path to data cube or to data and ct cubes
        cube = data_cube
        cube_basename = data_basename
    elif ct_cube is not None:  # user provided only path to ct cube and not to data cube
        cube = ct_cube
        cube_basename = ct_basename
    else:
        logger.error("Both (data and CT) cubes are empty")
        return 2

    # calculating common frame for printing cubes
    logger.info("Number of slices: " + str(cube.dimz))

    # get actual X and Y positions of bin centers, for two opposite corners, lowest and highest
    # i.e. assuming cube 5x5x5 with pixel_size 1 and slice_distance 1, we will have:
    # xmin_center, ymin_center, zmin_center = 0.5, 0.5, 0
    # xmax_center, ymax_center, zmax_center = 4.5, 4.5, 4
    xmin_center, ymin_center, zmin_center = cube.indices_to_pos([0, 0, 0])
    xmax_center, ymax_center, zmax_center = cube.indices_to_pos([cube.dimx - 1, cube.dimy - 1, cube.dimz - 1])

    logger.info("First bin center: {:10.2f} {:10.2f} {:10.2f} [mm]".format(xmin_center, ymin_center, zmin_center))
    logger.info("Last bin center : {:10.2f} {:10.2f} {:10.2f} [mm]".format(xmax_center, ymax_center, zmax_center))

    # get lowest corner of leftmost lowest bin
    xmin_lowest, ymin_lowest = xmin_center - 0.5 * cube.pixel_size, ymin_center - 0.5 * cube.pixel_size
    # get highest corner of highest rightmost bin
    xmax_highest, ymax_highest = xmax_center + 0.5 * cube.pixel_size, ymax_center + 0.5 * cube.pixel_size

    ct_cb = None
    data_cb = None
    ct_im = None
    data_im = None

    if args.sstart <= 0:
        logger.error("Invalid slice number {:d}. First slice must be 1 or greater.".format(args.sstart))
        return 1

    # if user hasn't provided number of first slice, then argument parsing method will set it to zero
    slice_start = args.sstart

    # user hasn't provided number of last slice, we assume then that the last one has number cube.dimz
    slice_stop = args.sstop
    if slice_stop is None:
        slice_stop = cube.dimz

    # Prepare figure and subplot (axis), they will stay the same during the loop
    fig = plt.figure()
    ax = fig.add_subplot(111, autoscale_on=False)

    # Set axis
    ax.set_xlabel("[mm]")
    ax.set_ylabel("[mm]")

    ax.set_aspect(1.0)
    for ax in fig.axes:
        ax.grid(True)

    ax.set_xlim(xmin_lowest, xmax_highest)
    ax.set_ylim(ymin_lowest, ymax_highest)

    minorLocator = MultipleLocator(cube.pixel_size)
    ax.xaxis.set_minor_locator(minorLocator)
    ax.yaxis.set_minor_locator(minorLocator)

    text_slice = ax.text(0.0, 1.01, "", size=8, transform=ax.transAxes)

    # loop over each slice
    for slice_id in range(slice_start, slice_stop + 1):

        output_filename = "{:s}_{:03d}.png".format(cube_basename, slice_id)

        slice_str = str(slice_id) + "/" + str(cube.dimz)
        if args.outputdir is not None:
            # use os.path.join() in order to add slash if it is missing.
            output_filename = os.path.join(args.outputdir, os.path.basename(output_filename))
        if args.verbosity == 0:
            print("Write slice number: " + slice_str)
        if args.verbosity > 0:
            logger.info("Write slice number: " + slice_str + " to " + output_filename)

        slice_str = "Slice #: {:3d}/{:3d}\nSlice position: {:6.2f} mm".format(slice_id,
                                                                              cube.dimz,
                                                                              cube.slice_to_z(slice_id - 1))
        text_slice.set_text(slice_str)

        if ct_cube is not None:
            ct_slice = ct_cube.cube[slice_id - 1, :, :]  # extract 2D slice from 3D cube

            # remove CT image from the current plot (if present) and replace later with new data
            if ct_im is not None:
                ct_im.remove()

            ct_im = ax.imshow(
                ct_slice,
                cmap=plt.cm.gray,
                interpolation='nearest',  # each pixel will get colour of nearest neighbour, useful when displaying
                #  dataset of lower resolution than the output image
                #  see http://matplotlib.org/examples/images_contours_and_fields/interpolation_methods.html
                origin="lower",  # ['upper' | 'lower']:
                # Place the [0,0] index of the array in the upper left or lower left corner of the axes.
                extent=[xmin_lowest, xmax_highest, ymin_lowest, ymax_highest]  # scalars (left, right, bottom, top) :
                # the location, in data-coordinates, of the lower-left and upper-right corners
            )

            # optionally add HU bar
            if args.HUbar and ct_cb is None:
                ct_cb = fig.colorbar(ct_im, ax=ax, ticks=arange(-1000, 3000, 200), orientation='horizontal')
                ct_cb.set_label('HU')

        if data_cube is not None:
            data_slice = data_cube.cube[slice_id - 1, :, :]  # extract 2D slice from 3D cube

            # remove data cube image from the current plot (if present) and replace later with new data
            if data_im is not None:
                data_im.remove()

            cmap1 = plt.cm.jet
            cmap1.set_under("k", alpha=0.0)  # seems to be broken! :-( -- Sacrificed goat, now working - David
            cmap1.set_over("k", alpha=0.0)
            cmap1.set_bad("k", alpha=0.0)  # Sacrificial knife here
            dmin = data_cube.cube.min()
            dmax = data_cube.cube.max()
            tmpdat = ma.masked_where(data_slice <= dmin, data_slice)  # Sacrifical goat

            # plot new data cube
            data_im = ax.imshow(
                tmpdat,
                cmap=cmap1,
                norm=colors.Normalize(vmin=0, vmax=dmax * 1.1, clip=False),
                alpha=0.7,
                interpolation='nearest',  # each pixel will get colour of nearest neighbour, useful when displaying
                #  dataset of lower resolution than the output image
                #  see http://matplotlib.org/examples/images_contours_and_fields/interpolation_methods.html
                origin="lower",  # ['upper' | 'lower']:
                # Place the [0,0] index of the array in the upper left or lower left corner of the axes.
                extent=[xmin_lowest, xmax_highest, ymin_lowest, ymax_highest]  # scalars (left, right, bottom, top) :
                # the location, in data-coordinates, of the lower-left and upper-right corners
            )

            if data_cb is None:
                data_cb = fig.colorbar(
                    data_im,
                    orientation='vertical',
                    shrink=0.8)
                if isinstance(data_cube, pt.LETCube):
                    data_cb.set_label("LET [keV/um]")
                elif isinstance(data_cube, pt.DosCube):
                    data_cb.set_label("Relative dose [%]")

        # by default savefil will produce 800x600 resolution, setting dpi=200 is increasing it to 1600x1200
        fig.savefig(output_filename, dpi=200)

    return 0


if __name__ == '__main__':
    logging.basicConfig()
    sys.exit(main(sys.argv[1:]))



#
#    Copyright (C) 2010-2016 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
The tripexecuter module provides functions for executing TRiP98 locally or remotely.
"""
# do not check this file for PEP8 compatibility
# flake8 complains about "E402 module level import not at top of file"
# flake8: noqa

from pytrip.tripexecuter.dosecube import DoseCube
from pytrip.tripexecuter.field import Field
from pytrip.tripexecuter.fieldcollection import FieldCollection
from pytrip.tripexecuter.tripexecuter import TripExecuter
from pytrip.tripexecuter.tripplan import TripPlan
from pytrip.tripexecuter.tripplancollection import TripPlanCollection
from pytrip.tripexecuter.tripvoi import TripVoi
from pytrip.tripexecuter.voi import Voi
from pytrip.tripexecuter.voicollection import VoiCollection
from pytrip.tripexecuter.rbehandler import RBEHandler, RBE

# from https://docs.python.org/3/tutorial/modules.html
# if a package's __init__.py code defines a list named __all__,
# it is taken to be the list of module names that should be imported when from package import * is encountered.
__all__ = ['DoseCube', 'Field', 'FieldCollection', 'TripExecuter', 'TripPlan', 'TripPlanCollection', 'TripVoi', 'Voi',
           'VoiCollection', 'RBEHandler', 'RBE']



#
#    Copyright (C) 2010-2016 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This module provides the Cube class, which is used by the CTX, DOS, LET and VDX modules.
A cube is a 3D object holding data, such as CT Hounsfield units, Dose- or LET values.
"""
import os
import re
import sys
import logging

import numpy as np

try:
    from dicom.dataset import Dataset, FileDataset
    from dicom import UID

    _dicom_loaded = True
except:
    _dicom_loaded = False

from pytrip.error import InputError, ModuleNotLoadedError

logger = logging.getLogger(__name__)


class Cube(object):
    """ Top level class for 3-dimensional data cubes used by e.g. DosCube, CtxCube and LETCube.
    The user should not use this class directly, but is instead referred to DosCube, CtxCube, LETCube classes
    and similar, as these inherits the Cube class and expose all attributes and methods from Cube.
    """
    def __init__(self, cube=None):
        if cube is not None:
            self.header_set = cube.header_set
            self.version = cube.version
            self.modality = cube.modality
            self.created_by = cube.created_by
            self.creation_info = cube.creation_info
            self.primary_view = cube.primary_view
            self.data_type = cube.data_type
            self.num_bytes = cube.num_bytes
            self.byte_order = cube.byte_order
            self.patient_name = cube.patient_name
            self.slice_dimension = cube.slice_dimension
            self.pixel_size = cube.pixel_size
            self.slice_distance = cube.slice_distance
            self.slice_number = cube.slice_number
            self.xoffset = cube.xoffset
            self.dimx = cube.dimx
            self.yoffset = cube.yoffset
            self.dimy = cube.dimy
            self.zoffset = cube.zoffset
            self.dimz = cube.dimz
            self.z_table = cube.z_table
            self.slice_pos = cube.slice_pos
            self._set_format_str()
            self._set_number_of_bytes()
            self.cube = np.zeros((self.dimz, self.dimy, self.dimx), dtype=cube.pydata_type)
        else:
            import getpass
            from pytrip import __version__ as _ptversion

            self.header_set = False
            self.version = "2.0"
            self.modality = "CT"
            self.created_by = getpass.getuser()
            self.creation_info = "Created with PyTRiP98 {:s}".format(_ptversion)
            self.primary_view = "transversal"  # e.g. transversal
            self.data_type = ""
            self.num_bytes = ""
            self.byte_order = "vms"  # aix or vms
            self.patient_name = ""
            self.slice_dimension = ""  # eg. 256 meaning 256x256 pixels.
            self.pixel_size = ""  # size in mm
            self.slice_distance = ""  # thickness of slice
            self.slice_number = ""  # number of slices in file.
            self.xoffset = 0
            self.dimx = ""  # number of pixels along x (e.g. 256)
            self.yoffset = 0
            self.dimy = ""
            self.zoffset = 0
            self.dimz = ""
            self.slice_pos = []
            self.z_table = False  # list of slice#,pos(mm),thickness(mm),tilt

    def __add__(self, other):
        c = type(self)(self)
        if Cube in other.__class__.__bases__:
            c.cube = other.cube + self.cube
        else:
            c.cube = self.cube + float(other)
        return c

    def __sub__(self, other):
        c = type(self)(self)
        if Cube in other.__class__.__bases__:
            c.cube = self.cube - other.cube
        else:
            c.cube = self.cube - float(other)
        return c

    def __mul__(self, other):
        c = type(self)(self)
        if Cube in other.__class__.__bases__:
            c.cube = other.cube * self.cube
        else:
            t = type(c.cube[0, 0, 0])
            c.cube = np.array(self.cube * float(other), dtype=t)
        return c

    def __div__(self, other):
        c = type(self)(self)
        if Cube in other.__class__.__bases__:
            c.cube = self.cube / other.cube
        else:
            t = type(c.cube[0, 0, 0])
            c.cube = np.array(self.cube / float(other), dtype=t)
        c.cube[np.isnan(c.cube)] = 0  # fix division by zero NaNs
        return c

    def is_compatible(self, other):
        """ Check if this Cube object is compatible in size and dimensions with 'other' cube.

        A cube object can be a CtxCube, DosCube, LETCube or similar object.
        Unlike check_compatibility(), this function compares itself to the other cube.

        :param Cube other: The other Cube object which will be checked compability with.
        :returns: True if compatibe.
        """
        return self.check_compatibility(self, other)

    @staticmethod
    def check_compatibility(a, b):
        """
        Simple comparison of cubes. if X,Y,Z dims are the same, and
        voxel sizes as well, then they are compatible. (Duck typed)

        See also the function is_compatible().

        :params Cube a: the first cube to be compared with the second (b).
        :params Cube b: the second cube to be compared with the first (a).

        """
        eps = 1e-5

        if a.dimx != b.dimx:
            return False
        elif a.dimy != b.dimy:
            return False
        elif a.dimz != b.dimz:
            return False
        elif (a.pixel_size - b.pixel_size) > eps:
            return False
        elif a.slice_distance != b.slice_distance:
            return False
        else:
            return True

    def indices_to_pos(self, indices):
        """ Translate index number of a voxel to real position in [mm], including any offsets.

        The z position is always following the slice positions.

        :params indices: tuple or list of integer indices (i,j,k) or [i,j,k]
        :returns: list of positions,including offsets, as a list of floats [x,y,z]
        """
        pos = [(indices[0] + 0.5) * self.pixel_size + self.xoffset,
               (indices[1] + 0.5) * self.pixel_size + self.yoffset,
               indices[2] * self.slice_distance + self.zoffset]
        return pos

    def slice_to_z(self, idx):
        """ Return z-position in [mm] of slice with index idx.

        :params int idx: index number of slice
        :returns: position of slice in [mm]
        """
        # TODO: out of bounds, return nearest neighbour. Update docstring.
        return idx * self.slice_distance + self.zoffset

    def pos_to_indices(self, pos):
        """ Translate a real x,y,z position in [mm] to voxel indices, including any offets.

        The z position is always following the slice positions.

        :params indices: tuple or list of float positions (x,y,z) or [x,y,z]
        :returns: list of positions,including offsets, as a list of floats [x,y,z]
        """
        indices = [int(pos[0] / self.pixel_size - self.xoffset / self.pixel_size),
                   int(pos[1] / self.pixel_size - self.yoffset / self.pixel_size),
                   int(pos[2] / self.slice_distance - self.zoffset / self.slice_distance)]
        return indices  # TODO: out of bounds, return nearest neighbour. Update docstring.

    def get_value_at_indice(self, idx):  # TODO: indice -> index (indece is wrong grammar)
        """ Retrieves the value of a voxel at index [i,j,k].

        :param [int*3] idx: list of integers descibing the i,j,k of the data cube.
        :returns: The voxel value at index[i,j,k]
        """
        # TODO: out of bounds, throw error.
        return self.cube[idx[2]][idx[1]][idx[0]]

    def get_value_at_pos(self, pos):
        """ Retrieves the value of a voxel at postion [x,y,z] in [mm], including any offsets.

        :param [int*3] idx: list of integers descibing the i,j,k of the data cube.
        :returns: The voxel value at index[i,j,k]
        """
        return self.get_value_at_indice(self.pos_to_indices(pos))

    def create_cube_from_equation(self, equation, center, limits, radial=True):
        """ Create Cube from a given equation.

        This function is currently out of order.

        """
        # TODO why eq not being used ?
        # eq = util.evaluator(equation)
        # TODO why data not being used ?
        # data = np.array(np.zeros((self.dimz, self.dimy, self.dimx)))
        x = np.linspace(0.5, self.dimx - 0.5, self.dimx) * self.pixel_size - center[0]
        y = np.linspace(self.dimx - 0.5, 0.5, self.dimx) * self.pixel_size - center[1]
        xv, yv = np.meshgrid(x, y)

    def load_from_structure(self, voi, preset=0, data_type=np.int16):
        """ Attaches/overwrites Cube.data based on a given Voi.

        Voxels within the structure are filled it with 'preset' value.
        Voxels outside the contour will be filled with Zeros.

        :param Voi voi: the volume of interest
        :param int preset: value to be assigned to the voxels within the contour.
        :param data_type: numpy data type, default is np.int16
        """
        data = np.array(np.zeros((self.dimz, self.dimy, self.dimx)), dtype=data_type)
        if preset != 0:
            for i_z in range(self.dimz):
                for i_y in range(self.dimy):
                    intersection = voi.get_row_intersections(self.indices_to_pos([0, i_y, i_z]))
                    if intersection is None:
                        break
                    if len(intersection) > 0:
                        k = 0
                        for i_x in range(self.dimx):
                            if self.indices_to_pos([i_x, 0, 0])[0] > intersection[k]:
                                k += 1
                                if k >= len(intersection):
                                    break
                            if k % 2 == 1:  # voxel is inside structure, if odd number of intersections.
                                data[i_z][i_y][i_x] = preset
        self.cube = data

    def create_empty_cube(self, value, dimx, dimy, dimz, pixel_size, slice_distance, slice_offset=0.0):
        """ Creates an empty Cube object.

        Values are stored as 2-byte integers.

        :param int16 value: integer value which will be assigned to all voxels.
        :param int dimx: number of voxels along x
        :param int dimy: number of voxels along y
        :param int dimz: number of voxels along z
        :param float pixel_size: size of each pixel (x == y) in [mm]
        :param float slice_distance: the distance between two slices (z) in [mm]
        :param float slice_offset: start position of the first slice in [mm] (default 0.0 mm)
        """
        self.dimx = dimx
        self.dimy = dimy
        self.dimz = dimz
        self.slice_number = dimz
        self.pixel_size = pixel_size
        self.slice_distance = slice_distance
        self.cube = np.ones((dimz, dimy, dimx), dtype=np.int16) * value
        self.slice_dimension = dimx
        self.num_bytes = 2
        self.data_type = "integer"
        self.pydata_type = np.int16
        self.slice_pos = [slice_distance * i + slice_offset for i in range(dimz)]
        self.header_set = True

    def override_cube_values(self, voi, value):
        """ Overwrites the Cube voxels within the given Voi with 'value'.

        Voxels within the structure are filled it with 'value'.
        Voxels outside the contour are not touched.

        :param Voi voi: the volume of interest
        :param value=0: value to be assigned to the voxels within the contour.
        """
        for i_z in range(self.dimz):
            for i_y in range(self.dimy):
                intersection = voi.get_row_intersections(self.indices_to_pos([0, i_y, i_z]))
                if intersection is None:
                    break
                if len(intersection) > 0:
                    k = 0
                    for i_x in range(self.dimx):
                        if self.indices_to_pos([i_x, 0, 0])[0] > intersection[k]:
                            k += 1
                            if k >= (len(intersection)):
                                break
                        if k % 2 == 1:  # voxel is inside structure, if odd number of intersections.
                            self.cube[i_z][i_y][i_x] = value

    def set_offset_cube_values(self, voi, value=0):
        """ Add 'value' to all voxels within the given Voi

        'value' is added to each voxel value within the given volume of interest.
        Voxels outside the volume of interest are not touched.

        :param Voi voi: the volume of interest
        :param value=0: value to be added to the voxel values within the contour.
        """
        for i_z in range(self.dimz):
            for i_y in range(self.dimy):
                intersection = voi.get_row_intersections(self.indices_to_pos([0, i_y, i_z]))
                if intersection is None:
                    break
                if len(intersection) > 0:
                    k = 0
                    for i_x in range(self.dimx):
                        if self.indices_to_pos([i_x, 0, 0])[0] > intersection[k]:
                            k += 1
                            if k >= (len(intersection)):
                                break
                        if k % 2 == 1:  # voxel is inside structure, if odd number of intersections.
                            self.cube[i_z][i_y][i_x] += value

    def write_trip_header(self, path):
        """ Write a TRiP98 formatted header file, based on the available meta data.

        :param path: fully qualified path, including file extention (.hed)
        """
        output_str = "version " + self.version + "\n"
        output_str += "modality " + self.modality + "\n"
        output_str += "created_by " + self.created_by + "\n"
        output_str += "creation_info " + self.creation_info + "\n"
        output_str += "primary_view " + self.primary_view + "\n"
        output_str += "data_type " + self.data_type + "\n"
        output_str += "num_bytes " + str(self.num_bytes) + "\n"
        output_str += "byte_order " + self.byte_order + "\n"
        if self.patient_name == "":
            self.patient_name = "Anonymous"
        output_str += "patient_name " + str(self.patient_name) + "\n"
        output_str += "slice_dimension " + str(self.slice_dimension) + "\n"
        output_str += "pixel_size " + str(self.pixel_size) + "\n"
        output_str += "slice_distance " + str(self.slice_distance) + "\n"
        output_str += "slice_number " + str(self.slice_number) + "\n"
        # output_str +=
        # "xoffset " + str(int(round(self.xoffset/self.pixel_size))) + "\n"
        output_str += "xoffset 0\n"
        output_str += "dimx " + str(self.dimx) + "\n"
        # output_str +=
        # "yoffset " + str(int(round(self.yoffset/self.pixel_size))) + "\n"
        output_str += "yoffset 0\n"
        output_str += "dimy " + str(self.dimy) + "\n"
        output_str += "zoffset 0\n"

        # """output_str += "zoffset " +
        # str(int(round(self.zoffset/self.slice_distance))) + "\n" """
        output_str += "dimz " + str(self.dimz) + "\n"
        output_str += "z_table no\n"

        # """if self.z_table is True:
        #     output_str += "z_table yes\n"
        #     output_str += "slice_no  position  thickness  gantry_tilt\n"
        #     for i in range(len(self.slice_pos)):
        #         output_str +=
        # "  %d\t%.4f\t%.4f\t%.4f\n"%(i+1,self.slice_pos[i],
        # self.slice_distance,0)"""
        with open(path, "w+") as f:
            f.write(output_str)

    def set_byteorder(self, endian=None):
        """Set/change the byte order of the data to be written to disk.

        Available options are:
        - 'little' vms, Intel style little-endian byte order.
        - 'big' aix, Motorola style big-endian byte order.
        - if unspecified, the native system dependent endianess is used.

        :param str endian: optional string containing the endianess.
        """
        if endian is None:
            endian = sys.byteorder
        if endian == 'little':
            self.byte_order = "vms"
        elif endian == 'big':
            self.byte_order = "aix"
        else:
            raise ValueError("set_byteorder error: unknown endian " + str(endian))

    def _set_format_str(self):
        """Set format string accoring to byte_order.
        """
        if (self.byte_order == "vms"):
            self.format_str = "<"
        elif (self.byte_order == "aix"):
            self.format_str = ">"
        self._set_number_of_bytes()

    def _set_number_of_bytes(self):
        """Set format_str and pydata_type according to num_bytes and data_type
        """
        if self.data_type == "integer":
            if self.num_bytes == 1:
                self.format_str += "b"
                self.pydata_type = np.int8
            if self.num_bytes == 2:
                self.format_str += "h"
                self.pydata_type = np.int16
            if self.num_bytes == 4:
                self.format_str += "i"
                self.pydata_type = np.int32
        elif self.data_type == "float":
            if self.num_bytes == 4:
                self.format_str += "f"
                self.pydata_type = np.float32
        elif self.data_type == "double":
            if self.num_bytes == 8:
                self.format_str += "d"
                self.pydata_type = np.double
        else:
            print("Format:", self.byte_order, self.data_type, self.num_bytes)
            raise IOError("Unsupported format.")

    def create_dicom_base(self):
        if _dicom_loaded is False:
            raise ModuleNotLoadedError("Dicom")
        if self.header_set is False:
            raise InputError("Header not loaded")
        meta = Dataset()
        meta.MediaStorageSOPClassUID = '1.2.840.10008.5.1.4.1.1.2'  # CT Image Storage
        meta.MediaStorageSOPInstanceUID = "1.2.3"
        meta.ImplementationClassUID = "1.2.3.4"
        meta.TransferSyntaxUID = UID.ImplicitVRLittleEndian  # Implicit VR Little Endian - Default Transfer Syntax
        ds = FileDataset("file", {}, file_meta=meta, preamble=b"\0" * 128)
        ds.PatientsName = self.patient_name
        ds.PatientID = "123456"
        ds.PatientsSex = '0'
        ds.PatientsBirthDate = '19010101'
        ds.SpecificCharacterSet = 'ISO_IR 100'
        ds.AccessionNumber = ''
        ds.is_little_endian = True
        ds.is_implicit_VR = True
        ds.SOPClassUID = '1.2.3'  # !!!!!!!!
        ds.SOPInstanceUID = '1.2.3'  # !!!!!!!!!!
        ds.StudyInstanceUID = '1.2.3'  # !!!!!!!!!!
        ds.FrameofReferenceUID = '1.2.3'  # !!!!!!!!!
        ds.StudyDate = '19010101'  # !!!!!!!
        ds.StudyTime = '000000'  # !!!!!!!!!!
        ds.PhotometricInterpretation = 'MONOCHROME2'
        ds.SamplesPerPixel = 1
        ds.ImageOrientationPatient = ['1', '0', '0', '0', '1', '0']
        ds.Rows = self.dimx
        ds.Columns = self.dimy
        ds.SliceThickness = str(self.slice_distance)

        ds.PixelSpacing = [self.pixel_size, self.pixel_size]
        return ds

    def merge(self, cube):
        self.cube = np.maximum(self.cube, cube.cube)

    def merge_zero(self, cube):
        self.cube[self.cube == 0] = cube.cube[self.cube == 0]

    @staticmethod
    def discover_file(file_name):
        """
        Checks if file_name exists in the filesystem. If yes, gives back its path.
        If not, checks if gzipped file with same name exists.
        If gzipped file exists, gives back it path, otherwise returns None.

        :param file_name: File name or path
        :return: file_name, file_name + ".gz" or None
        """
        gzip_file_name = file_name + ".gz"
        if os.path.exists(file_name):
            logger.debug("File " + file_name + " exists")
            return file_name
        elif os.path.exists(gzip_file_name):
            logger.debug("File " + gzip_file_name + " exists")
            return gzip_file_name
        else:
            logger.error("Both " + file_name + " and " + gzip_file_name + " do not exists")
            return None

    @classmethod
    def parse_path(cls, path_name):
        """
        Parse path_name which can have form of: bare name (i.e. TST001), plain file (TST001.hed or TST001.ctx),
        gzipped file (TST001.hed.gz or TST001.ctx.gz) or some other name. Calculates plain file names of
        header and data file. Extension of data file is extracted from the class from which this method was called
        (.ctx for CtxCube, .dos for DosCube, .dosemlet.dos for LetCube). In case of non-parseable data None is returned.

        >>> from pytrip import CtxCube
        >>> CtxCube.parse_path("frodo.hed")
        ('frodo.hed', 'frodo.ctx')
        >>> CtxCube.parse_path("baggins.ctx")
        ('baggins.hed', 'baggins.ctx')
        >>> CtxCube.parse_path("mordor")
        ('mordor.hed', 'mordor.ctx')
        >>> CtxCube.parse_path("legolas.hed.gz")
        ('legolas.hed', 'legolas.ctx')
        >>> CtxCube.parse_path("gimli.son.of.gloin.ctx.gz")
        ('gimli.son.of.gloin.hed', 'gimli.son.of.gloin.ctx')
        >>> CtxCube.parse_path("bilbo.dos")
        (None, None)
        >>> pt.CtxCube.parse_path("/home/pytrip/patient.ctx")
        ('/home/pytrip/patient.hed', '/home/pytrip/patient.ctx')
        >>> from pytrip import DosCube
        >>> DosCube.parse_path("bilbo.dos")
        ('bilbo.hed', 'bilbo.dos')
        >>> DosCube.parse_path("baggins.ctx")
        (None, None)
        >>> from pytrip import LETCube
        >>> LETCube.parse_path("aragorn.dosemlet.dos")
        ('aragorn.dosemlet.hed', 'aragorn.dosemlet.dos')
        >>> LETCube.parse_path("aragorn.dosemlet")
        (None, None)
        >>> LETCube.parse_path("aragorn")
        ('aragorn.dosemlet.hed', 'aragorn.dosemlet.dos')
        >>> LETCube.parse_path("aragorn.ctx")
        (None, None)

        :param path_name: path to header file, data file or basename path (path to file without extension).
        Path can be absolute or relative. It can also lead to gzipped files.
        :return: pair of filenames for header and data
        """

        logger.info("Parsing " + path_name)

        # this method is intended to be used by subclasess of Cube (i.e. DosCube, LETCube, CtxCube)
        # let us ensure first if these subclassess define file extensions corresponding to header and data files

        # checking if current class has data extension attribute, to compare with data file extension
        if 'data_file_extension' in cls.__dict__:
            class_data_file_extension = cls.data_file_extension.lower()
            logger.debug("class data filename extension: " + class_data_file_extension)
        else:  # class without extension, we assume in such case file is given without extension
            error_msg = "Class " + str(cls) + " doesn't have data_file_extension field"
            logger.error(error_msg)
            raise NotImplementedError(error_msg)

        # checking if current class has header extension attribute, to compare with header file extension
        if 'header_file_extension' in cls.__dict__:
            class_header_file_extension = cls.header_file_extension.lower()
            logger.debug("class data filename extension: " + class_header_file_extension)
        else:  # class without extension, we assume in such case file is given without extension
            error_msg = "Class " + str(cls) + " doesn't have header_file_extension field"
            logger.error(error_msg)
            raise NotImplementedError(error_msg)

        #  path can end with .gz or not (be it archive or not). Let us check it
        is_path_gzipped = path_name.endswith(".gz")
        logger.debug("is file gzipped ? " + str(is_path_gzipped))

        # strip .gz from path (if present)
        uncompressed_path_name = path_name
        if is_path_gzipped:
            uncompressed_path_name = os.path.splitext(path_name)[0]
        logger.debug("uncompressed filename: " + uncompressed_path_name)

        # we have several options:
        # 1. file with header extension (i.e. blabla.hed or blabla.dosemlet.hed)
        # 2. file with data extension (i.e. blabla.ctx, blabla.dos., blabla.dosemlet.dos)
        # 3. file without extension (i.e. blabla)
        # 4. file with unknown extensions (i.e. blabla.mp3)
        # Options 1-3 can be parsed, option 4 will give us empty result

        if uncompressed_path_name.endswith(class_header_file_extension):  # header file
            logger.debug("header file ")
            header_file = uncompressed_path_name
            basename_end_index = uncompressed_path_name.rfind(class_header_file_extension)
            cube_filename = uncompressed_path_name[:basename_end_index] + class_data_file_extension
        elif uncompressed_path_name.endswith(class_data_file_extension):  # cube file
            logger.debug("cube file ")
            basename_end_index = uncompressed_path_name.rfind(class_data_file_extension)
            header_file = uncompressed_path_name[:basename_end_index] + class_header_file_extension
            cube_filename = uncompressed_path_name
        elif not os.path.splitext(uncompressed_path_name)[1]:  # file without extension
            logger.debug("file without extension ")
            header_file = uncompressed_path_name + class_header_file_extension
            cube_filename = uncompressed_path_name + class_data_file_extension
        else:   # a problem (i.e. uknown exception)
            logger.debug("a problem (unknown extension?) with path " + path_name)
            header_file, cube_filename = None, None

        return header_file, cube_filename

    def read_trip_header(self, content):
        i = 0
        self.header_set = True
        content = content.split('\n')
        has_ztable = False
        while i < len(content):
            if re.match("version", content[i]) is not None:
                self.version = content[i].split()[1]
            if re.match("modality", content[i]) is not None:
                self.modality = content[i].split()[1]
            if re.match("created_by", content[i]) is not None:
                self.created_by = content[i].replace("created_by ", "", 1)
                self.created_by = self.created_by.rstrip()
            if re.match("creation_info", content[i]) is not None:
                self.creation_info = content[i].replace("creation_info ", "", 1)
                self.creation_info = self.creation_info.rstrip()
            if re.match("primary_view", content[i]) is not None:
                self.primary_view = content[i].split()[1]
            if re.match("data_type", content[i]) is not None:
                self.data_type = content[i].split()[1]
            if re.match("num_bytes", content[i]) is not None:
                self.num_bytes = int(content[i].split()[1])
            if re.match("byte_order", content[i]) is not None:
                self.byte_order = content[i].split()[1]
            if re.match("patient_name", content[i]) is not None:
                self.patient_name = content[i].split()[1]
            if re.match("slice_dimension", content[i]) is not None:
                self.slice_dimension = int(content[i].split()[1])
            if re.match("pixel_size", content[i]) is not None:
                self.pixel_size = float(content[i].split()[1])
            if re.match("slice_distance", content[i]) is not None:
                self.slice_distance = float(content[i].split()[1])
            if re.match("slice_number", content[i]) is not None:
                self.slice_number = int(content[i].split()[1])
            if re.match("xoffset", content[i]) is not None:
                self.xoffset = int(content[i].split()[1])
            if re.match("yoffset", content[i]) is not None:
                self.yoffset = int(content[i].split()[1])
            if re.match("zoffset", content[i]) is not None:
                self.zoffset = int(content[i].split()[1])
            if re.match("dimx", content[i]) is not None:
                self.dimx = int(content[i].split()[1])
            if re.match("dimy", content[i]) is not None:
                self.dimy = int(content[i].split()[1])
            if re.match("dimz", content[i]) is not None:
                self.dimz = int(content[i].split()[1])
            if re.match("slice_no", content[i]) is not None:
                self.slice_pos = [float(j) for j in range(self.slice_number)]
                has_ztable = True
                i += 1
                for j in range(self.slice_number):
                    self.slice_pos[j] = float(content[i].split()[1])
                    i += 1
            i += 1
        self.zoffset *= self.slice_distance
        if has_ztable is not True:
            self.slice_pos = [float(j) for j in range(self.slice_number)]
            for i in range(self.slice_number):
                self.slice_pos[i] = self.zoffset + i * self.slice_distance
        self._set_format_str()
        self._set_number_of_bytes()

    def read(self, path):
        """ Reads both TRiP98 data and its associated header into the Cube object.

        :param str path: Path to filename to be read, file extention may be given but is not neccesary.
        """
        self.read_trip_data_file(path)

    def read_trip_header_file(self, path):  # TODO: could be made private? #126
        """ Reads a header file, accepts also if suffix is missing, or if file
        is .gz compressed. User can thus specify:
        tst001
        tst001.hed
        tst001.hed.gz

        and in any case it will attempt to load
        tst001.hed
        or
        tst001.hed.gz

        Note, first the un-zipped files will be attempted to read.
        Should these not exist, then the .gz are attempted.

        However, if the .hed.gz file was explicitly stated,
        then this file will also
        be loaded, even if a .hed is available.
        """

        # extract header name (blabla.hed) from path
        # we get blabla.hed even if path equals to blabla.hed.gz, see docs in Cube.parse_path function
        header_file_name, _ = self.parse_path(path)

        # figure out which file exists on disc: *.hed or *.hed.gz and get its path
        header_file_path = self.discover_file(header_file_name)

        # sanity check
        if header_file_path is not None:
            logger.info("Reading header file" + header_file_path)
        else:
            raise IOError("Could not find file " + path)

        # load plain of gzipped file
        if header_file_path.endswith(".gz"):
            import gzip
            fp = gzip.open(header_file_path, "rt")
        else:
            fp = open(header_file_path, "rt")
        content = fp.read()
        fp.close()

        # fill self with data
        self.read_trip_header(content)
        self._set_format_str()
        logger.debug("Format string:" + self.format_str)

    def read_trip_data_file(self, path, multiply_by_2=False):  # TODO: could be made private? #126
        """Read TRiP98 formatted data.

        Accepts path in similar way as read_trip_header_file().
        If header file was not previously loaded, it will be attepted first.

        Due to an issue in VIRTUOS, sometimes DosCube data have been reduced with a factor of 2.
        Setting multiply_by_2 to True, will restore the true values, in this case.

        :param path: Path to TRiP formatted data.
        :param multiply_by_2: The data read will automatically be multiplied with a factor of 2.
        """
        # extract header and data file name from path
        header_file_name, data_file_name = self.parse_path(path)

        # fill header data if self.header is empty
        if self.header_set is False:
            header_file_path = self.discover_file(header_file_name)

            # sanity check
            if header_file_path is not None:
                logger.info("Reading header file" + header_file_path)
                self.read_trip_header_file(header_file_path)
            else:
                raise IOError("Could not find file " + path)

        # raise exception if reading header failed
        if self.header_set is False:
            raise InputError("Header file not loaded")

        # preparation
        data_dtype = np.dtype(self.format_str)
        data_count = self.dimx * self.dimy * self.dimz

        # figure out path to data file on disk, might be gzipped or not
        data_file_path = self.discover_file(data_file_name)

        # load data from data file (gzipped or not)
        logger.info("Opening file: " + path)
        if data_file_path.endswith('.gz'):
            import gzip
            with gzip.open(data_file_path, "rb") as f:
                s = f.read(data_dtype.itemsize * data_count)
                cube = np.frombuffer(s, dtype=data_dtype, count=data_count)
        else:
            cube = np.fromfile(data_file_path, dtype=data_dtype)

        if self.byte_order == "aix":
            logger.info("AIX big-endian data.")
            # byteswapping is not needed anymore, handled by "<" ">" in dtype

        # sanity check
        logger.info("Cube data points : {:d}".format(len(cube)))
        if len(cube) != self.dimx * self.dimy * self.dimz:
            logger.error("Header size and cube size dont match.")
            logger.error("Cube data points : {:d}".format(len(cube)))
            logger.error("Header says      : {:d} = {:d} * {:d} * {:d}".format(
                self.dimx * self.dimy * self.dimz,
                self.dimx,
                self.dimy,
                self.dimz))
            raise IOError("Header data and dose cube size are not consistent.")

        cube = np.reshape(cube, (self.dimz, self.dimy, self.dimx))
        if multiply_by_2:
            logger.warning("Cube was previously rescaled to 50%. Now multiplying with 2.")
            cube *= 2
        self.cube = cube

    def set_data_type(self, type):
        """ Sets the data type for the TRiP98 header files.

        :param numpy.type type: numpy type, e.g. np.uint16
        """
        if type is np.int8 or type is np.uint8:
            self.data_type = "integer"
            self.num_bytes = 1
        elif type is np.int16 or type is np.uint16:
            self.data_type = "integer"
            self.num_bytes = 2
        elif type is np.int32 or type is np.uint32:
            self.data_type = "integer"
            self.num_bytes = 4
        elif type is np.float:
            self.data_type = "float"
            self.num_bytes = 4
        elif type is np.double:
            self.data_type = "double"
            self.num_bytes = 8

    def read_dicom_header(self, dcm):
        """ Creates the header metadata for this Cube class, based on a given Dicom object.

        :param Dicom dcm: Dicom object which will be used for generating the header data.
        """
        if _dicom_loaded is False:
            raise ModuleNotLoadedError("Dicom")
        ds = dcm["images"][0]
        self.version = "1.4"
        self.created_by = "pytrip"
        self.creation_info = "Created by PyTRiP98;"
        self.primary_view = "transversal"
        self.set_data_type(type(ds.pixel_array[0][0]))
        self.patient_name = ds.PatientsName
        self.slice_dimension = int(ds.Rows)  # should be changed ?
        self.pixel_size = float(ds.PixelSpacing[0])
        self.slice_distance = abs(
            float(dcm["images"][0].ImagePositionPatient[2]) - float(dcm["images"][1].ImagePositionPatient[2]))
        self.slice_number = len(dcm["images"])
        self.xoffset = float(ds.ImagePositionPatient[0])
        self.dimx = int(ds.Rows)
        self.yoffset = float(ds.ImagePositionPatient[1])
        self.dimy = int(ds.Columns)
        self.zoffset = float(ds.ImagePositionPatient[2])
        self.dimz = len(dcm["images"])
        self.z_table = True
        self.set_z_table(dcm)
        self.set_byteorder()
        self._set_format_str()
        self.header_set = True

    def set_z_table(self, dcm):
        """ Creates the slice lookup table based on a given Dicom object.
        :param Dicom dcm: dicom object provided by pydicom.
        """
        self.slice_pos = []
        for i in range(len(dcm["images"])):
            self.slice_pos.append(float(dcm["images"][i].ImagePositionPatient[2]))

    def write_trip_data(self, path):
        """ Writes the binary data cube in TRiP98 format to a file.

        Type is specified by self.pydata_type and self.byte_order attributes.

        :param str path: Full path including file extention.
        """
        cube = np.array(self.cube, dtype=self.pydata_type)
        if self.byte_order == "aix":
            cube = cube.byteswap()
        cube.tofile(path)
        return
        # TODO after return - probably forgotten
        # f = open(path, "wb+")
        # out = ""
        # _format = self.format_str[0] + self.format_str[1] * self.dimx
        # # i = 0 TODO why not used ?
        # for image in self.cube:
        #     out = ""
        #     for line in image:
        #         out += pack(_format, *line)
        #     f.write(out)
        # f.close()



#
#    Copyright (C) 2010-2016 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This module provides the top-level Rst class which can handle accelerator control files generated by TRiP98.
It also provides the SubMachine class which treats individual energy layers.
"""
import numpy as np
import dicom
from pytrip.file_parser import parse_to_var


class Rst:
    """ This class handles raster scan data, which are accelerator control files in GSI format.
    Raster scan data are stored in .rst file, and describe the amount of particles going into
    each spot in each energy layer. Each energy layer is called a 'submachine'.
    """
    def __init__(self):
        self.var_dict = {"rstfile": "rstfile",
                         "sistable": "sistable",
                         "patient_id": "patient_name",
                         "projectile": "projectile",
                         "gantryangle": "gantryangle",
                         "couchangle": "couchangle",
                         "#submachines": "submachines",
                         "bolus": "bolus",
                         "ripplefilter": "ripplefilter",
                         "mass": "mass",
                         "charge": "charge",
                         "#particles": "particles"}
        self.machines = []

    def get_submachines(self):
        """
        :returns: A list of submachines.
        """
        return self.machines

    def calculate_total_energy(self):  # TODO: not implemented
        """
        Not implemented.
        """
        return

    def load_dicom(self, path):  # TODO: not implemented, load_dicom() -> read_from_dicom()
        """ Load a Dicom file from 'path'

        Currently, this function merely stores the dicom data into self.data.
        No interpretation is done.

        :param str path: Full path to Dicom file.
        """
        self.data = dicom.read_file(path)

    def read(self, path):
        """ Load and parse a raster scan (.rst) file.

        :param str path: Full path to the file to be loaded, including file extension.
        """
        with open(path, mode='r') as f:
            data = f.read()
        data = data.split("\n")
        out, i = parse_to_var(data, self.var_dict, "submachine#")

        for key, item in out.items():
            setattr(self, key, item)
        if hasattr(self, "bolus"):
            self.bolus = float(self.bolus)
        for machine in range(int(self.submachines)):
            submachine = SubMachine()
            i = submachine._load_submachine(data, i)
            self.machines.append(submachine)

    def get_stepsize(self):
        """ Returns the distance between each spot in the first energy plane.

        Most likely the distance will be the same in all planes.

        :returns: Distancce between spots in [mm]. If no submachines are found, None is returned.
        """
        if len(self.machines) > 0:
            return self.machines[0].stepsize
        return None

    def get_min_max(self):
        """ Retrieve the largest and smallest x,y position found in all energy layers.

        :returns: A list of four values in [x_min,x_max,y_min,y_max] in [mm].
        """
        min_x = None
        max_x = None
        min_y = None
        max_y = None
        for submachine in self.machines:
            min_max = submachine.get_raster_min_max()
            if min_x is None or min_x > min_max[0]:
                min_x = min_max[0]
            if max_x is None or max_x > min_max[1]:
                max_x = min_max[1]
            if min_y is None or min_y > min_max[2]:
                min_y = min_max[2]
            if max_y is None or max_y > min_max[3]:
                max_y = min_max[3]
        return [min_x, max_x, min_y, max_y]

    def save_random_error_rst(self, path, sigma):
        # TODO why out here, not being used
        # out = self.generate_random_error_rst(sigma)

        with open(path, "wb+") as fp:
            fp.write("rstfile %s\n" % getattr(self, "rstfile"))
            fp.write("sistable %s\n" % getattr(self, "sistable"))
            fp.write("patient_id %s\n" % getattr(self, "patient_name"))
            fp.write("machine# 0\n")
            fp.write("projectile %s\n" % getattr(self, "projectile"))
            fp.write("charge %s\n" % getattr(self, "charge"))
            fp.write("mass %s\n" % getattr(self, "mass"))
            fp.write("gantryangle %s\n" % getattr(self, "gantryangle"))
            fp.write("couchangle %s\n" % getattr(self, "couchangle"))
            fp.write("stereotacticcoordinates\n")
            fp.write("bolus %.0f\n" % getattr(self, "bolus"))
            fp.write("ripplefilter %s\n" % getattr(self, "ripplefilter"))
            fp.write("#submachines %s\n" % getattr(self, "submachines"))
            fp.write("#particles %s\n" % getattr(self, "particles"))
            for machine in self.machines:
                machine.save_random_error_machine(fp, sigma)

    def generate_random_error_rst(self, sigma):
        out = []
        for submachine in self.machines:
            out.append(submachine.generate_random_error_machine(sigma))


class SubMachine:
    def __init__(self):
        self.raster_points = []

    def get_raster_points(self):
        return self.raster_points

    def get_raster_min_max(self):
        """ Returns the smallest and largest x and y positions for this energy layer.
        :returns: a list of four elements [min_x, max_x, min_y, max_y]
        """
        min_x = min(self.raster_points, key=lambda x: x[0])[0]
        min_y = min(self.raster_points, key=lambda x: x[1])[1]
        max_x = max(self.raster_points, key=lambda x: x[0])[0]
        max_y = max(self.raster_points, key=lambda x: x[1])[1]
        return [min_x, max_x, min_y, max_y]

    def get_raster_grid(self):
        min_max = self.get_raster_min_max()
        zero = [-int(min_max[0] / self.stepsize[0]), -int(min_max[2] / self.stepsize[1])]
        size = [(min_max[1] - min_max[0]) / self.stepsize[0] + 1, (min_max[3] - min_max[2]) / self.stepsize[1] + 1]
        grid = np.zeros(size)
        rasterpoints = np.array(self.raster_points)
        grid[np.array(rasterpoints[:, 0] / self.stepsize[0] + zero[0], 'uint8'), np.array(
            rasterpoints[:, 1] / self.stepsize[1] + zero[1], 'uint8')] = rasterpoints[:, 2]
        return grid

    def generate_random_error_machine(self, sigma):
        """ Generates a new set of raster points, including distributed setup errors with a Gaussian sigma.

        This function is used to simulate positioning errors.
        The existing raster field is shifted using a Gaussian distribution. This corresponds to blurring the
        raster scan file.
        Spot weights below 2000 particles are truncated.

        :param float sigma: sigma of the Gaussian blur to be applied [mm]
        """
        # prepare a grid which covers (xmin,ymin) to (xmax,ymax)
        min_max = self.get_raster_min_max()
        zero = [-int(min_max[0] / self.stepsize[0]),
                -int(min_max[2] / self.stepsize[1])]  # start position of steps [i_min,j_min], may be negative
        size = [(min_max[1] - min_max[0]) / self.stepsize[0] + 1,
                (min_max[3] - min_max[2]) / self.stepsize[1] + 1]  # no of steps to reach opposite side [i_len,j_len]
        grid = np.zeros((size))  # 2d grid with steop positions

        # Fill the new grid with raster values from the original raster points
        rasterpoints = np.array(self.raster_points)  # cast raster points into numpy array
        grid[np.array(rasterpoints[:, 0] / self.stepsize[0] + zero[0], 'uint8'),
             np.array(rasterpoints[:, 1] / self.stepsize[1] + zero[1], 'uint8')] = rasterpoints[:, 2]

        # extend grid size a bit to cover positions going beyond the edges when applying an offset.
        # determine the new size:
        size2 = [size[0] + 4 * int(sigma / self.stepsize[0]),  # width
                 size[1] + 4 * int(sigma / self.stepsize[1])]
        zero2 = [zero[0] + 2 * int(sigma / self.stepsize[0]),  # start positions
                 zero[1] + 2 * int(sigma / self.stepsize[1])]

        # make new grid with new size
        outgrid = np.zeros(size2)

        # calculate a few offset positions from a Gaussian.
        # offset holds offset indices. Mostly this will be just [-1,0,1] for x and y.
        offset = np.array([[x, y]
                           for x in range(-int(sigma / self.stepsize[0]), int(sigma / self.stepsize[0]) + 1)
                           for y in range(-int(sigma / self.stepsize[1]), int(sigma / self.stepsize[1]) + 1)])

        # Translate integer offsets to actual positions, squared.
        lengths = (offset[:, 0] * self.stepsize[0]) ** 2 + \
                  (offset[:, 1] * self.stepsize[1]) ** 2

        gauss = np.exp(-0.5 * lengths / sigma ** 2)  # Array shaping a Gaussian distribution from lengths
        gauss = gauss / np.sum(gauss)  # Normalize it

        # extend number of offsets
        offset[:, 0] = offset[:, 0] + 2 * int(sigma / self.stepsize[0])  # x
        offset[:, 1] = offset[:, 1] + 2 * int(sigma / self.stepsize[1])  # y

        # For each position, modify the spot intensity to be a sum of the offsets.
        for i in range(len(offset)):
            o = offset[i]
            outgrid[o[0]:o[0] + size[0], o[1]:o[1] + size[1]] += gauss[i] * grid

        # store new grid in l which will be returned
        _l = []
        for i in range(len(outgrid)):
            for j in range(len(outgrid[0])):  # run along y indices
                # only append to result grid _l, if we have a reasonable number of particles in this spot.
                if outgrid[i, j] > 2000:  # spot threshold is 2000 particles
                    _l.append([(i - zero2[0]) * self.stepsize[0],
                               (j - zero2[1]) * self.stepsize[1],
                               outgrid[i, j]])
        return _l

    def save_random_error_machine(self, fp, sigma):
        """ Generates and stores a single energy layer where Gaussian blur has been applied.

        :param fp: file pointer
        :param float sigma: sigma of the Gaussian blur to be applied [mm]
        """
        rasterpoints = np.array(self.generate_random_error_machine(sigma))
        fp.write("submachine# %d %.2f %d %.1f\n" % (self.idx_energy, self.energy, self.idx_focus, self.focus))
        fp.write("#particles %.5E %.5E %.5E\n" % (np.min(rasterpoints[:, 2]), np.max(rasterpoints[:, 2]),
                                                  np.sum(rasterpoints[:, 2])))
        fp.write("stepsize %.0f %.0f\n" % (self.stepsize[0], self.stepsize[1]))
        fp.write("#points %d\n" % (len(rasterpoints)))
        for point in rasterpoints:
            fp.write("%d %d %.5E\n" % (point[0], point[1], point[2]))

    def _load_submachine(self, data, i):
        """ Loads a single submachine from 'data'.

        :param [str] data: list of lines containing the .rst data.
        :param i: line number to start from
        :returns: new line number, when parsing has finished.
        """
        items = data[i].split()
        self.idx_energy = int(items[1])
        self.energy = float(items[2])
        self.idx_focus = int(items[3])
        self.focus = float(items[4])

        while True:
            i += 1
            line = data[i]
            if line.find("stepsize") > -1:
                items = line.split()
                self.stepsize = [float(items[1]), float(items[2])]
            if line.find("#points") > -1:
                self.points = int(line.split()[1])
                break
            if line.find("#particles") > -1:
                self.min_particles = line.split()[1]
                self.max_particles = line.split()[2]
                self.total_particles = line.split()[3]
        i += 1
        for i in range(i, i + self.points):
            items = data[i].split()
            self.raster_points.append([float(items[0]), float(items[1]), float(items[2])])
        return i + 1



#
#    Copyright (C) 2010-2016 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
TODO: documentation here.
"""
import os
import shutil
import tarfile
import copy
import uuid

import numpy as np
from subprocess import Popen, PIPE

try:
    import paramiko
except:
    pass

from pytrip.dos import DosCube
from pytrip.let import LETCube
from pytrip.vdx import VdxCube
from pytrip.error import InputError
from pytrip.res.point import get_basis_from_angles, angles_to_trip
import pytriplib
import logging

logger = logging.getLogger(__name__)


class TripExecuter(object):
    def __init__(self, images, rbe=None):
        logger.debug("Initializing TripExecuter()")
        self.images = images
        self.rbe = rbe
        self.listeners = []
        self.trip_bin_path = "TRiP98"  # where TRiP98 is installed, if not accessible in /usr/local/bin or similar
        self.logfile_stdout = "trip98.stdout"
        self.logfile_stderr = "trip98.stderr"
        self.rsakey_local_path = "~/.ssh/id_rsa"
        self._runtrip = True  # set this to False for a dry run without TRiP98 (for testing purposes)
        self.remote_dir = "."  # remote directory where all will be run

    def delete_workspace(self):
        shutil.rmtree(self.path)

    def cube_in_other_cube(self, outer, inner):
        m = np.max(inner.cube)
        idx = np.where(inner.cube == m)
        idx = [idx[0][0], idx[1][0], idx[2][0]]
        a = outer.cube[idx[0], idx[1]]
        b = inner.cube[idx[0], idx[1]]
        res = a > b
        if True in res[0:idx[2]] and True in res[idx[2]:-1]:
            return True
        return False

    def analyse_cube(self):
        keys = self.projectiles.keys()

        p1 = self.projectiles[keys[0]]
        p2 = self.projectiles[keys[1]]

        cube1 = p1["target_dos"]
        cube2 = p2["target_dos"]

        if not self.cube_in_other_cube(cube1, cube2):
            temp = p1
            p1 = p2
            p2 = temp
            cube1 = p1["target_dos"]
            cube2 = p2["target_dos"]
            self.execute_order = [keys[0], keys[1]]
            self.split_proj_key = keys[1]

        else:
            self.execute_order = [keys[1], keys[0]]
            self.split_proj_key = keys[0]

        target_cube = copy.deepcopy(cube1)
        shadow_cubes = []
        for i, field in enumerate(p1["fields"]):
            d = DosCube(cube1)

            basis = get_basis_from_angles(field.get_gantry(), field.get_couch())
            basis = basis[0]
            basis = np.array([basis[0] / cube1.pixel_size, basis[1] / cube1.pixel_size,
                              basis[2] / cube1.slice_distance])
            basis /= np.max(np.abs(basis))

            d.cube = pytriplib.create_field_shadow(
                cube1.cube,
                cube2.cube,
                np.array(basis, dtype=np.double))
            target_cube -= d
            shadow_cubes.append(d)

        target_cube.cube[target_cube.cube < 0] = 0
        cube2.cube = cube2.cube + target_cube.cube
        # ~ cube2.cube = pytriplib.extend_cube(cube2.cube)
        cube1.cube = cube1.cube - target_cube.cube
        if len(p1["fields"]) == 2:
            a = self.execute_order.pop(1)
            b = self.projectile_dose_level[a]
            self.execute_order.append(a + str(1))
            self.execute_order.append(a + str(2))
            self.projectile_dose_level[a + str(1)] = b
            self.projectile_dose_level[a + str(2)] = b

    def execute(self, plan, callback=None):
        """
        Executes the plan object using TRiP98.
        """
        self.plan = plan
        self.callback = callback
        self.ini_execute()
        if not self.mult_proj:
            self.execute_simple()
        else:
            self.execute_mult_proj()
        if os.path.exists(os.path.join(self.path, self.plan_name) + ".bio.dos") and self.plan.get_out_bio_dose():
            self.plan.load_dose(os.path.join(self.path, self.plan_name) + ".bio.dos", "bio", self.target_dose)
        if os.path.exists(os.path.join(self.path, self.plan_name) + ".phys.dos") and self.plan.get_out_phys_dose():
            self.plan.load_dose(os.path.join(self.path, self.plan_name) + ".phys.dos", "phys", self.target_dose)
        if os.path.exists(os.path.join(self.path, self.plan_name) + ".phys.dos") and self.plan.get_out_dose_mean_let():
            self.plan.load_let(os.path.join(self.path, self.plan_name) + ".dosemlet.dos")
        self.finish()

    def ini_execute(self):
        self.split_plan(self.plan)
        self.plan_name = self.plan.get_name().replace(" ", "_")
        self.working_path = os.path.expandvars(self.plan.get_working_dir())
        if not hasattr(self, "folder_name"):
            self.folder_name = str(uuid.uuid4())
        self.path = os.path.join(self.working_path, self.folder_name)
        self.prepare_folder()
        self.convert_files_to_voxelplan()

    def execute_simple(self):
        """
        Simple execution of TRiP, called by self.execute() when self.mult_proj is not set.
        """
        logger.debug("In execute_simple mode")
        self.create_trip_exec_simple()
        self.run_trip()

    def execute_mult_proj(self):
        self.analyse_cube()
        i = 3

        for a in range(i):
            for projectile in self.execute_order:
                self.calculate_rest_dose(projectile)
                if a == i - 1:
                    self.create_trip_exec_mult_proj(projectile, first=(a is 0))
                else:
                    self.create_trip_exec_mult_proj(projectile, last=False, first=(a is 0))
                self.run_trip()

                self.split_fields(projectile)
        self.post_process()

    def create_trip_exec_simple(self, no_output=False):
        """
        Generates the .exec script and stores it into self.trip_exec.
        """
        logger.info("Generating the trip_exec script...")
        oar_list = self.oar_list
        targets = self.targets
        fields = self.plan.get_fields()

        projectile = fields[0].get_projectile()
        output = []
        output.extend(self.create_exec_header())
        output.extend(self.create_exec_load_data_files(projectile))
        output.extend(self.create_exec_field(fields))
        output.extend(self.create_exec_oar(oar_list))
        dosecube = None
        if len(targets) > 1:
            dosecube = DosCube(self.images)
            for i, voi in enumerate(targets):
                temp = DosCube(self.images)
                dose_level = int(voi.get_dose() / self.target_dose * 1000)
                if dose_level == 0:
                    dose_level = -1
                temp.load_from_structure(voi.get_voi().get_voi_data(), dose_level)
                if i == 0:
                    dosecube = temp * 1
                else:
                    dosecube.merge_zero(temp)
            dosecube.cube[dosecube.cube == -1] = int(0)

        if not self.plan.get_target_dose_cube() is None:
            dosecube = self.plan.get_target_dose_cube()

        if dosecube is not None:
            if not no_output:
                dosecube.write(os.path.join(self.path, "target_dose.dos"))
            output.extend(self.create_exec_plan(incube="target_dose"))
        else:
            output.extend(self.create_exec_plan())
        if self.plan.get_optimize():
            output.extend(self.create_exec_opt())

        name = self.plan_name
        output.extend(self.create_exec_output(name, fields))
        out = "\n".join(output) + "\nexit\n"
        self.trip_exec = out

    def create_exec_output(self, name, fields, last=True):
        output = []
        window = self.plan.get_window()
        window_str = ""
        if len(window) is 6:
            window_str = " window(%.2f,%.2f,%.2f,%.2f,%.2f,%.2f) " % (window[0], window[1], window[2], window[3],
                                                                      window[4], window[5])

        if self.plan.get_out_phys_dose() is True:
            output.append('dose "' + name + '." /calculate  alg(' + self.plan.get_dose_algorithm() + ')' + window_str +
                          '  field(*) write')
        if last:
            if self.plan.get_out_bio_dose() is True:
                output.append('dose "' + name + '." /calculate ' + window_str + ' bioalgorithm(' +
                              self.plan.get_bio_algorithm() + ') biological norbe write')
            if self.plan.get_out_dose_mean_let() is True:
                output.append('dose "' + name + '." /calculate ' + window_str + ' field(*) dosemeanlet write')
            if self.plan.get_out_field() is True and self.plan.get_optimize() is True:
                for i, field in enumerate(fields):
                    output.append('field %d /write file(%s.rst) reverseorder ' % (i + 1, field.get_name()))
                    field.set_rasterfile(self.working_path + '//' + self.folder_name + '//' + field.get_name())
        return output

    def create_exec_plan(self, incube=""):
        output = []
        plan = "plan / dose(%.2f) " % self.target_dose
        if not self.plan.get_target_tissue_type() == "":
            plan += "target(%s) " % (self.plan.get_target_tissue_type())
        if not self.plan.get_res_tissue_type() == "":
            plan += "residual(%s) " % (self.plan.get_res_tissue_type())
        if not incube == "":
            plan += "incube(%s) " % incube
        output.append(plan)
        return output

    def create_exec_opt(self):
        output = []
        opt = "opt / field(*) "
        opt += self.plan.get_opt_method() + " "
        opt += "iterations(" + str(self.plan.get_iterations()) + ") "
        opt += "dosealgorithm(" + self.plan.get_dose_algorithm() + ") "
        opt += "" + self.plan.get_opt_princip() + " "
        opt += "geps(" + str(self.plan.get_geps()) + ") "
        opt += "eps(" + str(self.plan.get_eps()) + ") "
        opt += "optalgorithm(" + self.plan.get_opt_algorithm() + ") "
        opt += "bioalgorithm(" + self.plan.get_bio_algorithm() + ") "

        output.append(opt)
        return output

    def create_exec_header(self):
        output = []
        output.append("time / on")
        output.append("sis  * /delete")
        output.append("hlut * /delete")
        output.append("ddd  * /delete")
        output.append("dedx * /delete")
        output.append('dedx "' + self.plan.dedx_file + '" / read')
        output.append('hlut "' + self.plan.hlut_file + '" / read')
        output.append("scancap / offh2o(1.709) rifi(3) bolus(0.000) " "minparticles(5000) path(none)")
        output.append("random 1000")
        return output

    def create_exec_load_data_files(self, projectile):
        output = []
        ddd_folder = self.plan.get_ddd_folder()
        spc_folder = self.plan.get_spc_folder()
        sis_file = self.plan.get_sis_file()
        if len(ddd_folder) > 0:
            output.append('ddd "%s/*" / read' % ddd_folder)
        if len(spc_folder) > 0:
            output.append('spc "%s/*" / read' % spc_folder)
        if len(sis_file) > 0:
            output.append('sis "%s" / read' % sis_file)
        else:
            output.append('sis / dummy')
        if not (len(spc_folder) > 0 or len(ddd_folder) > 0):
            if projectile == "C":
                output.append('ddd "$TRIP98/DATA/DDD/12C/RF3MM/12C*" / read')
                output.append('spc "$TRIP98/DATA/SPC/12C/RF3MM/12C*" / read')
            elif projectile == "H":
                output.append('ddd "$TRIP98/DATA/DDD/1H/RF3MM/1H*" / read')
                output.append('spc "$TRIP98/DATA/SPC/1H/RF3MM/1H*" / read')
            elif projectile == "O":
                output.append('ddd "$TRIP98/DATA/DDD/16O/RF3MM/16O*" / read')
                output.append('spc "$TRIP98/DATA/SPC/16O/RF3MM/16O*" / read')
            elif projectile == "Ne":
                output.append('ddd "$TRIP98/DATA/DDD/20Ne/RF3MM/20Ne*" / read')
                output.append('spc "$TRIP98/DATA/SPC/20Ne/RF3MM/20Ne*" / read')
        output.append("ct " + self.plan_name + " /read")

        output.append("voi " + self.plan_name + "  /read")
        output.append("voi * /list")

        if not self.plan.get_target_tissue_type() == "":
            rbe = self.rbe.get_rbe_by_name(self.plan.get_target_tissue_type())
            output.append("rbe '%s' /read" % (rbe.get_path()))
            if not self.plan.get_res_tissue_type() == "" and \
                    not self.plan.get_res_tissue_type() \
                    == self.plan.get_target_tissue_type():
                rbe = self.rbe.get_rbe_by_name(self.plan.get_target_tissue_type())
                output.append("rbe %s /read" % (rbe.get_path()))

        return output

    def create_exec_field(self, fields):
        output = []
        if self.plan.get_optimize():
            for i, val in enumerate(fields):
                field = "field " + str(i + 1) + " / new "
                field += "fwhm(%d) " % (val.get_fwhm())
                raster = val.get_rasterstep()
                if not raster[0] is 0 and not raster[1] is 0:
                    field += "raster(%.2f,%.2f) " % (raster[0], raster[1])
                gantry, couch = angles_to_trip(val.get_gantry(), val.get_couch())
                field += "couch(" + str(couch) + ") "
                field += "gantry(" + str(gantry) + ") "
                target = val.get_target()
                if len(val.get_target()) is not 0:
                    field += "target(%.1f,%.1f,%.1f) " % (target[0], target[1], target[2])
                if val.get_doseextension() > 0.0001:
                    field += "doseext(" + str(val.get_doseextension()) + ") "
                field += "contourext(" + str(val.get_contourextension()) + ") "
                if val.get_zsteps() > 0.001:
                    field += "zsteps(" + str(val.get_zsteps()) + ") "
                field += 'proj(' + val.get_projectile() + ')'
                output.append(field)
        else:
            for i, val in enumerate(fields):
                field = 'field 1 /read file(' + val.get_rasterfile() + '.rst)'
                field += "fwhm(%d) " % (val.get_fwhm())
                raster = val.get_rasterstep()
                if not raster[0] is 0 and not raster[1] is 0:
                    field += "raster(%.2f,%.2f) " % (raster[0], raster[1])
                gantry, couch = angles_to_trip(val.get_gantry(), val.get_couch())
                field += "couch(" + str(couch) + ") "
                field += "gantry(" + str(gantry) + ") "
                if len(val.get_target()) is not 0:
                    field += "target(" + val.get_target() + ") "
                if val.get_doseextension() > 0.0001:
                    field += "doseext(" + str(val.get_doseextension()) + ") "
                field += "contourext(" + str(val.get_contourextension()) + ") "
                if val.get_zsteps() > 0.001:
                    field += "zsteps(" + str(val.get_zsteps()) + ") "
                field += 'proj(' + val.get_projectile() + ')'
                output.append(field)
        return output

    def create_exec_oar(self, oar_list):
        output = []
        for oar in oar_list:
            output.append("voi " + oar.get_name().replace(" ", "_") + " / maxdosefraction(" + str(
                oar.get_max_dose_fraction()) + ") oarset")
        return output

    def split_fields(self, proj):
        if self.split_proj_key not in self.projectiles.keys():
            return
        name = os.path.join(self.path, self.plan_name) + "_" + self.projectiles[proj]["name"]
        path = os.path.join(name + ".phys.dos")
        temp = DosCube()
        temp.read(path)
        temp.version = "2.0"

        self.rest_dose = self.target_dos - temp

        p = self.projectiles[self.split_proj_key]
        p["target_dos"].cube = pytriplib.extend_cube(p["target_dos"].cube)
        if len(p["fields"]) == 2:
            temp.cube = (temp.cube < self.projectiles[proj]["target_dos"].cube) * \
                self.projectiles[proj]["target_dos"].cube + \
                (temp.cube > self.projectiles[proj]["target_dos"].cube) * temp.cube
            dose = self.target_dos - temp
            field1 = p["fields"][0]
            field2 = p["fields"][1]
            d1 = DosCube(temp)
            d2 = DosCube(temp)
            center = pytriplib.calculate_dose_center(p["target_dos"].cube)

            dose.cube[dose.cube < 0] = 0

            temp.cube *= self.target_dos.cube > 0

            basis = get_basis_from_angles(field1.get_gantry(), field1.get_couch())[0]
            basis = np.array([basis[0] / dose.pixel_size, basis[1] / dose.pixel_size, basis[2] / dose.slice_distance])
            basis = basis / np.max(np.abs(basis)) * .5
            d1.cube = pytriplib.create_field_shadow(
                dose.cube,
                temp.cube,
                np.array(basis, dtype=np.double))
            basis = get_basis_from_angles(field2.get_gantry(), field2.get_couch())[0]
            basis = np.array([basis[0] / dose.pixel_size, basis[1] / dose.pixel_size, basis[2] / dose.slice_distance])
            basis /= np.max(np.abs(basis))
            d2.cube = pytriplib.create_field_shadow(
                dose.cube,
                temp.cube,
                np.array(basis, dtype=np.double))
            a = d2.cube > d1.cube
            b = d2.cube < d1.cube
            d2.cube = p["target_dos"].cube * a
            d1.cube = p["target_dos"].cube * b

            rest = p["target_dos"].cube * ((a + b) < 1)

            b = pytriplib.split_by_plane(rest, center, basis)
            d1.cube += b
            d2.cube += rest - b
            self.plan.add_dose(d1, "H1")
            self.plan.add_dose(d2, "H2")

            self.projectiles[field1.get_projectile() + str(1)] = {
                "target_dos": d1,
                "fields": [field1],
                "name": field1.get_projectile() + str(1),
                "projectile": field1.get_projectile()
            }
            self.projectiles[field2.get_projectile() + str(2)] = {
                "target_dos": d2,
                "fields": [field2],
                "name": field2.get_projectile() + str(2),
                "projectile": field2.get_projectile()
            }
            del self.projectiles[self.split_proj_key]

    def calculate_rest_dose(self, proj):
        self.rest_dose = copy.deepcopy(self.target_dos)
        for k, projectile in self.projectiles.items():
            if k == proj:
                continue
            name = os.path.join(self.path, self.plan_name) + "_" + projectile["name"]
            path = os.path.join(name + ".phys.dos")
            if os.path.exists(path):
                temp = DosCube()
                temp.read(path)
                temp = temp
                self.rest_dose.cube = self.rest_dose.cube - temp.cube
                # ~ self.rest_dose.cube[self.rest_dose.cube<0] = 0

    def post_process(self):
        phys_dose = None
        bio_dose = None
        dose_mean_let = None
        temp_dos = None
        for projectile in self.projectiles:
            name = os.path.join(self.path, self.plan_name) + "_" + self.projectiles[projectile]["name"]
            factor = float(self.projectile_dose_level[projectile]) / 1000
            factor = 1
            if os.path.exists(name + ".bio.dos") and self.plan.get_out_bio_dose():
                path = os.path.join(name + ".bio.dos")
                temp = DosCube()
                temp.read(path)
                temp *= factor
                if self.mult_proj:
                    self.plan.add_dose(temp, "bio_%s" % projectile)
                if bio_dose is None:
                    bio_dose = temp
                else:
                    bio_dose += temp

            if os.path.exists(name + ".phys.dos") and self.plan.get_out_phys_dose():
                path = os.path.join(name + ".phys.dos")
                temp_dos = DosCube()
                temp_dos.read(path)
                temp_dos *= factor
                if self.mult_proj:
                    self.plan.add_dose(temp_dos, "phys_%s" % projectile)
                if phys_dose is None:
                    phys_dose = temp_dos
                else:
                    phys_dose += temp_dos
            if os.path.exists(name + ".dosemlet.dos") and self.plan.get_out_dose_mean_let():

                path = os.path.join(name + ".dosemlet.dos")
                temp = LETCube()
                temp.read(path)
                if not self.mult_proj:
                    dose_mean_let = temp
                else:
                    if dose_mean_let is None:
                        dose_mean_let = temp * temp_dos
                    else:
                        dose_mean_let = dose_mean_let + temp * temp_dos
        out_path = os.path.join(self.path, self.plan_name)
        if bio_dose is not None:
            bio_dose.write(out_path + ".bio.dos")
        if phys_dose is not None:
            phys_dose.write(out_path + ".phys.dos")
        if dose_mean_let is not None:
            if self.mult_proj:
                dose_mean_let /= phys_dose

            dose_mean_let.write(out_path + ".hed")

    def split_plan(self, plan=None):
        self.targets = []
        self.oar_list = []

        dose = 0
        for voi in self.plan.get_vois():
            if voi.is_oar():
                self.oar_list.append(voi)
            if voi.is_target():
                self.targets.append(voi)
                if voi.get_dose() > dose:
                    dose = voi.get_dose()
        if not len(self.targets):
            raise InputError("No targets")
        if not len(self.plan.get_fields()):
            raise InputError("No fields")
        self.target_dose = dose
        if plan is None:
            plan = self.plan
        proj = []
        self.projectile_dose_level = {}
        for field in plan.fields:
            if field.get_projectile() not in proj:
                proj.append(field.get_projectile())
                self.projectile_dose_level[field.get_projectile()] = 0

        if len(proj) > 1:
            self.mult_proj = True
        else:
            self.mult_proj = False

        if self.mult_proj:
            self.projectiles = {}
            for field in plan.fields:
                if field.get_projectile() not in self.projectiles.keys():
                    self.projectiles[field.get_projectile()] = {
                        "target_dos": DosCube(self.images),
                        "fields": [field],
                        "name": field.get_projectile(),
                        "projectile": field.get_projectile()
                    }
                else:
                    self.projectiles[field.get_projectile()]["fields"].append(field)

            self.target_dos = DosCube(self.images)

            for i, voi in enumerate(self.targets):
                temp = DosCube(self.images)
                voi_dose_level = int(voi.get_dose() / dose * 1000)
                temp.load_from_structure(voi.get_voi().get_voi_data(), 1)
                for projectile, data in self.projectiles.items():
                    dose_percent = self.plan.get_dose_percent(projectile)
                    if not voi.get_dose_percent(projectile) is None:
                        dose_percent = voi.get_dose_percent(projectile)
                    proj_dose_lvl = int(voi.get_dose() / self.target_dose * dose_percent * 10)
                    if self.projectile_dose_level[projectile] < proj_dose_lvl:
                        self.projectile_dose_level[projectile] = proj_dose_lvl
                    if proj_dose_lvl == 0:
                        proj_dose_lvl = -1
                    if i == 0:
                        data["target_dos"] = temp * proj_dose_lvl
                    else:
                        data["target_dos"].merge_zero(temp * proj_dose_lvl)
                if i == 0:
                    self.target_dos = temp * voi_dose_level
                else:
                    self.target_dos.merge_zero(temp * voi_dose_level)
            for projectile, data in self.projectiles.items():
                data["target_dos"].cube[data["target_dos"].cube == -1] = int(0)
                self.plan.add_dose(data["target_dos"], "target_%s" % projectile)
            self.rest_dose = copy.deepcopy(self.target_dos)

    def add_log_listener(self, listener):
        self.listeners.append(listener)

    def log(self, txt):
        txt = txt.replace("\n", "")
        for l in self.listeners:
            l.write(txt)

    def prepare_folder(self):
        self.filepath = self.path + self.plan_name
        if os.path.exists(self.path):
            pass
            # shutil.rmtree(self.path)
        else:
            os.makedirs(self.path)

    def run_trip(self):
        if self.plan.remote:
            self.run_trip_remote()
        else:
            self.run_trip_local()

    def run_trip_remote(self):
        logger.info("Run TRiP98 in REMOTE mode.")
        # self.create_remote_run_file()
        self.compress_files()

        self.copy_files_to_server()

        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        # If no password is supplied, try to look for a private key
        if self.plan.get_password() is "" or None:
            rsa_keypath = os.path.expanduser(self.rsakey_local_path)
            if not os.path.isfile(rsa_keypath):
                # login with provided username + empty password
                try:
                    ssh.connect(self.plan.get_server(), username=self.plan.get_username(), password="")
                except:
                    logger.error("Cannot connect to " + self.plan.get_server())
                    logger.error("Check username, password or key in " + self.rsakey_local_path)
                    raise
            else:
                # login with provided username + private key
                rsa_key = paramiko.RSAKey.from_private_key_file(rsa_keypath)
                try:
                    ssh.connect(self.plan.get_server(), username=self.plan.get_username(), pkey=rsa_key)
                except:
                    logger.error("Cannot connect to " + self.plan.get_server())
                    logger.error("Check username and your key in " + self.rsakey_local_path)
                    raise
        else:
            # login with provided username + password
            ssh.connect(self.plan.get_server(), username=self.plan.get_username(), password=self.plan.get_password())

        if not self._runtrip:
            norun = "echo "
        else:
            norun = ""

        path_tgz = os.path.join(self.remote_dir, "temp.tar.gz")  # remote place where temp.tar.gz is stored
        rrun_dir = os.path.join(self.remote_dir, self.folder_name)  # remote run dir

        commands = ["cd " + self.remote_dir + ";" + "tar -zxvf " + path_tgz,
                    "cd " + rrun_dir + ";" + norun + "bash -lc '" + self.trip_bin_path + " < plan.exec '",
                    "cd " + self.remote_dir + ";" + "tar -zcvf " + path_tgz + " " + rrun_dir,
                    "cd " + self.remote_dir + ";" + "rm -r " + rrun_dir]

        logger.debug("Write stdout and stderr to " + os.path.join(self.working_path, self.folder_name))
        fp_stdout = open(os.path.join(self.working_path, self.folder_name, self.logfile_stdout), "w")
        fp_stderr = open(os.path.join(self.working_path, self.folder_name, self.logfile_stderr), "w")

        for cmd in commands:
            logger.debug("Execute on remote: " + cmd)
            self.log(cmd)
            stdin, stdout, stderr = ssh.exec_command(cmd)
            answer_stdout = stdout.read()
            answer_stderr = stderr.read()
            logger.debug("Remote answer stdout:" + answer_stdout)
            logger.debug("Remote answer stderr:" + answer_stderr)
            fp_stdout.write(answer_stdout)
            fp_stderr.write(answer_stderr)
            self.log(answer_stdout)
        ssh.close()
        fp_stdout.close()
        fp_stderr.close()

        self.copy_back_from_server()
        self.decompress_data()

    def run_trip_local(self):
        """
        Runs TRiP98 on local computer.
        """
        logger.info("Run TRiP98 in LOCAL mode.")
        logger.debug("Write stdout and stderr to " + os.path.join(self.working_path, self.folder_name))
        fp_stdout = open(os.path.join(self.working_path, self.folder_name, self.logfile_stdout), "w")
        fp_stderr = open(os.path.join(self.working_path, self.folder_name, self.logfile_stderr), "w")

        os.chdir("%s" % (self.path))

        if not self._runtrip:  # for testing
            norun = "echo "
        else:
            norun = ""
        p = Popen([norun + self.trip_bin_path], stdout=PIPE, stdin=PIPE)

        p.stdin.write(self.trip_exec)
        p.stdin.flush()
        while (True):
            retcode = p.poll()  # returns None while subprocess is running
            answer_stdout = p.stdout.readline()
            answer_stderr = p.stderr.readline()
            logger.debug("Remote answer stdout:" + answer_stdout)
            logger.debug("Remote answer stderr:" + answer_stderr)
            fp_stdout.write(answer_stdout)
            fp_stderr.write(answer_stderr)
            self.log(answer_stdout)
            if retcode is not None:  # runs until process stops
                break  # TODO: check return code
        os.chdir("..")

        fp_stdout.close()
        fp_stderr.close()

    def finish(self):
        pass

    def create_trip_exec_mult_proj(self, projectile, no_output=False, last=True, first=True):

        fields = self.projectiles[projectile]["fields"]
        oar_list = self.oar_list

        output = []
        output.extend(self.create_exec_header())
        output.extend(self.create_exec_load_data_files(self.projectiles[projectile]["projectile"]))
        output.extend(self.create_exec_field(fields))
        output.extend(self.create_exec_oar(oar_list))

        dosecube = copy.deepcopy(self.projectiles[projectile]["target_dos"])

        if not no_output:
            if hasattr(self, "rest_dose"):
                dosecube.cube = np.array(
                    (self.rest_dose.cube >= dosecube.cube) * dosecube.cube +
                    (self.rest_dose.cube < dosecube.cube) * self.rest_dose.cube,
                    dtype=np.int16)
                dosecube.cube[dosecube.cube < 0] = 0
                # ~ self.plan.add_dose(self.rest_dose,"rest")
                if not first:
                    a = (self.rest_dose.cube - dosecube.cube) * (dosecube.cube > 0)
                    dosecube.cube += a * dosecube.cube / 500

            dosecube.write(os.path.join(self.path, "target_dose_%s.dos" % self.projectiles[projectile]["name"]))

        self.projectile_dose_level[projectile] = np.max(dosecube.cube)
        output.extend(self.create_exec_plan(incube="target_dose_%s" % self.projectiles[projectile]["name"]))
        if self.plan.get_optimize() is True:
            output.extend(self.create_exec_opt())

        name = self.plan_name + "_" + self.projectiles[projectile]["name"]
        output.extend(self.create_exec_output(name, fields))
        out = "\n".join(output) + "\nexit\n"
        self.trip_exec = out

    def convert_files_to_voxelplan(self):
        out_path = os.path.join(self.path, self.plan_name)
        ctx = self.images
        ctx.patient_name = self.plan_name
        ctx.write(os.path.join(out_path + ".ctx"))
        structures = VdxCube(ctx)
        structures.version = "2.0"
        liste = []
        area = 0
        for voi in self.plan.get_vois():
            voxelplan_voi = voi.get_voi().get_voi_data()

            if voi.is_target():
                mn, mx = voxelplan_voi.get_min_max()
                area_temp = (mx[0] - mn[0]) * (mx[1] - mn[1]) * (mx[2] - mn[2])
                if area_temp > area:
                    area = area_temp
                    liste.insert(0, voxelplan_voi)
                else:
                    liste.append(voxelplan_voi)
                voxelplan_voi.type = '1'
            else:
                liste.append(voxelplan_voi)
                voxelplan_voi.type = '0'
        for voxelplan_voi in liste:
            structures.add_voi(voxelplan_voi)
        structures.write_to_trip(out_path + ".vdx")

    def compress_files(self):
        """
        Builds the tar.gz from what is found in self.path.
        """

        logger.debug("Compressing files in " + self.path)
        self.save_exec(os.path.join(self.working_path, self.folder_name, "plan.exec"))
        with tarfile.open(os.path.join(self.working_path, self.folder_name + ".tar.gz"), "w:gz") as tar:
            tar.add(self.path, arcname=self.folder_name)

    def set_plan(self, plan):
        self.plan = plan
        self.plan_name = self.plan.get_name().replace(" ", "_")

    def save_exec(self, path):
        """
        Writes the .exec script to disk.
        """
        self.split_plan()
        if self.mult_proj:
            path1 = path.replace(".exec", "")
            for projectile in self.projectiles:
                self.create_trip_exec(projectile, True)
                with open(path1 + "_" + projectile + ".exec", "wb+") as fp:
                    fp.write(self.trip_exec)
        else:
            with open(path, "wb+") as fp:
                fp.write(self.trip_exec)

    def save_data(self, path):
        out_path = path
        ctx = self.images
        ctx.patient_name = self.plan_name
        ctx.write(os.path.join(out_path + ".ctx"))
        structures = VdxCube(ctx)
        structures.version = "2.0"
        for voi in self.plan.get_vois():
            voxelplan_voi = voi.get_voi().get_voi_data()
            structures.add_voi(voxelplan_voi)
            if voi.is_target():
                voxelplan_voi.type = '1'
            else:
                voxelplan_voi.type = '0'
        structures.write_to_trip(out_path + ".vdx")

    def get_transport(self):
        transport = paramiko.Transport((self.plan.get_server(), 22))
        transport.connect(username=self.plan.get_username(), password=self.plan.get_password())
        return transport

    def copy_files_to_server(self):
        """
        Copies the generated tar.gz file to the remote server.
        """
        logger.debug("Copy tar.gz to server:" + self.path + ".tar.gz -> " +
                     os.path.join(self.remote_dir, "temp.tar.gz"))
        transport = self.get_transport()
        sftp = paramiko.SFTPClient.from_transport(transport)
        sftp.put(localpath=self.path + ".tar.gz",
                 remotepath=os.path.join(self.remote_dir, 'temp.tar.gz'))
        sftp.close()
        transport.close()

    def run_ssh_command(self, cmd):
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(self.plan.get_server(), username=self.plan.get_username(), password=self.plan.get_password())
        self.parent.write_to_log("Run Trip\n")
        stdin, stdout, stderr = ssh.exec_command(cmd)
        ssh.close()

    def copy_back_from_server(self):
        transport = self.get_transport()
        sftp = paramiko.SFTPClient.from_transport(transport)
        sftp.get(os.path.join(self.remote_dir, 'temp.tar.gz'), self.path + ".tar.gz")
        sftp.close()
        transport.close()

    def decompress_data(self):
        output_folder = self.path
        if os.path.exists(output_folder):
            shutil.rmtree(output_folder)
        with tarfile.open(self.path + ".tar.gz", "r:gz") as tar:
            tar.extractall(self.working_path)

    def clean_up(self):
        f = "%s" % (self.path + ".tar.gz")
        if os.path.exists(f):
            os.remove(f)
        shutil.rmtree("%s" % self.path)

    def visualize_data(self):
        pass



#
#    Copyright (C) 2010-2016 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This module provides the LETCube for handling LET data.
"""
import numpy as np
from pytrip.cube import Cube
import pytriplib


class LETCube(Cube):
    """ This class handles LETCubes.

    It is similar to DosCubes and CtxCubes, but is intended to hold LET data.
    The object has build-in methods to read and write the LET data cubes,
    calculate the LET-volume historgrams, and write these to disk.
    It is inherited from Cube, which contains many additional methods and attributes.
    """

    data_file_extension = ".dosemlet.dos"
    header_file_extension = ".dosemlet.hed"

    def __init__(self, cube=None):
        super(LETCube, self).__init__(cube)

    def write(self, path):
        """Write the LETCube and its header to a file with the filename 'path'.

        :param str path: path to header file, data file or basename (without extension)

        """

        header_file, let_file = self.parse_path(path_name=path)
        self.write_trip_header(header_file)
        self.write_trip_data(let_file)

    def get_max(self):
        """ Returns the largest value in the LETCube.

        :returns: the largets value found in the in the LETCube.
        """
        return np.amax(self.cube)

    def calculate_lvh(self, voi):
        """ Calculate a LET-volume historgram.

        :param Voi voi: The volume of interest, in the form of a Voi object.
        :returns: A tuple containing
                  - lvh: the LET-volume histogram
                  - min_lvh: array of LET values below 2%
                  - max_lvh: array of LET values above 98%
                  - area: TODO - what is this?
        """
        pos = 0
        size = np.array([self.pixel_size, self.pixel_size, self.slice_distance])
        lv = np.zeros(3000)
        for i in range(self.dimz):
            pos += self.slice_distance
            slice = voi.get_slice_at_pos(pos)
            if slice is not None:
                lv += pytriplib.calculate_lvh_slice(self.cube[i], np.array(slice.contour[0].contour), size)

        cubes = sum(lv)
        lvh = np.cumsum(lv[::-1])[::-1] / cubes
        min_let = np.where(lvh >= 0.98)[0][-1]
        max_let = np.where(lvh <= 0.02)[0][0]
        area = cubes * size[0] * size[1] * size[2] / 1000
        mean = np.dot(lv, range(0, 3000)) / cubes
        return lvh, min_let, max_let, mean, area

    def write_lvh_to_file(self, voi, path):
        """ Write the LET-volume historgram to a file.

        :param Voi voi: The volume of interest, n the form of a Voi object.
        :param str path: Full path of file to be written.
        """
        lvh, _, _, _, _ = self.calculate_lvh(voi)
        print(lvh.shape)
        output = ""
        # TODO fix following line
        # lvh has shape (3000,0) and contains only let, not vol
        # see calculate_lvh() method
        for vol, let in zip(lvh[0], lvh[1]):
            output += "%.4e\t%.4e\n" % (vol, let)
        f = open(path + ".dvh", "w+")
        f.write(output)
        f.close()



#!/usr/bin/env python
#
#    Copyright (C) 2010-2016 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Script for converting a DICOM file to TRiP98 / Voxelplan style files.
"""
import sys
import logging
import argparse

import pytrip as pt

logger = logging.getLogger(__name__)


def main(args=sys.argv[1:]):
    """ Main function for dicom2trip.py
    """
    # parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("dicom_folder", help="location of folder with DICOM files", type=str)
    parser.add_argument("ctx_basename", help="basename of output file in TRiP98 format", type=str)
    parser.add_argument("-v", "--verbosity", action='count', help="increase output verbosity", default=0)
    parser.add_argument('-V', '--version', action='version', version=pt.__version__)
    args = parser.parse_args(args)

    basename = args.ctx_basename

    if args.verbosity == 1:
        logger.setLevel('INFO')
    if args.verbosity > 1:
        logger.setLevel('DEBUG')

    # import DICOM
    dcm = pt.dicomhelper.read_dicom_folder(args.dicom_folder)
    c = pt.CtxCube()
    c.read_dicom(dcm)

    vdx_cube = pt.VdxCube(cube=c)

    try:
        vdx_cube.read_dicom(dcm)
        vdx_cube.write_trip_data(basename + ".vdx")
    except:
        logger.warning("No contour information found in DICOM")
        pass

    c.write_trip_header(basename + pt.CtxCube.header_file_extension)
    c.write_trip_data(basename + pt.CtxCube.data_file_extension)

    return 0


if __name__ == '__main__':
    logging.basicConfig()
    sys.exit(main(sys.argv[1:]))



#
#    Copyright (C) 2010-2016 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This module provides the DensityCube class and some special class to find robust angles for treament
based on a quality index factor defined in http://dx.doi.org/10.3109/0284186X.2015.1067720.
"""
import gc
import os
from math import sin, cos
from multiprocessing import Process, Queue
from functools import cmp_to_key

import numpy as np
from scipy import interpolate

from pytrip.cube import Cube
from pytrip.res.point import get_basis_from_angles
import pytriplib


def cmp_sort(a, b):
    """ Sorting key. If gantry angle is equal, then sort by couch angle.
    """
    if a["gantry"] == b["gantry"]:
        return a["couch"] - b["couch"]
    return a["gantry"] - b["gantry"]


class DensityProjections:
    """
    Functions here were mosly used buy for the publication http://dx.doi.org/10.3109/0284186X.2015.1067720
    """
    def __init__(self, cube):
        """ TODO: Documentation
        """
        self.cube = cube

    def calculate_quality_grid(self, voi, gantry, couch, calculate_from=0, stepsize=1.0, avoid=[], gradient=True):
        """ TODO: Documentation
        """
        l = self.calculate_quality_list(voi, gantry, couch, calculate_from, stepsize, avoid=avoid, gradient=gradient)
        l = sorted(l, key=cmp_to_key(cmp_sort))
        grid_data = []
        for x in l:
            grid_data.append(x["data"][0])
        l = np.reshape(grid_data, (len(gantry), len(couch)))
        return l

    def calculate_quality_list(self, voi, gantry, couch, calculate_from=0, stepsize=1.0, avoid=[], gradient=True):
        """ TODO: Documentation
        """
        q = Queue(32767)
        process = []
        d = voi.get_voi_cube()
        d.cube = np.array(d.cube, dtype=np.float32)
        voi_cube = DensityProjections(d)
        result = []
        for gantry_angle in gantry:
            p = Process(
                target=self.calculate_angle_quality_thread,
                args=(voi, gantry_angle, couch, calculate_from, stepsize, q, avoid, voi_cube, gradient))
            p.start()
            p.deamon = True
            process.append(p)
            if len(process) > 2:
                tmp = q.get()
                result.append(tmp)
                for p in process:
                    if not p.is_alive():
                        process.remove(p)
        while not len(result) == len(gantry) * len(couch):
            tmp = q.get()
            result.append(tmp)
        return result

    def calculate_best_angles(self, voi, fields, min_dist=20, gantry_limits=[-90, 90], couch_limits=[0, 359], avoid=[]):
        """ TODO: Documentation
        """
        grid = self.calculate_quality_list(
            voi,
            range(gantry_limits[0], gantry_limits[1], 20),
            range(couch_limits[0], couch_limits[1], 20),
            avoid=avoid)
        grid = sorted(grid, key=lambda x: x["data"][0])
        best_angles = []
        i = 0
        while len(best_angles) < fields or i >= len(grid):
            is_ok = True
            point = np.array([grid[i]["gantry"], grid[i]["couch"]])
            for angle in best_angles:
                if np.linalg.norm(angle - point) < 2 * min_dist:
                    is_ok = False
            if is_ok:
                best_angles.append(self.optimize_angle(voi, point[0], point[1], 20, 3, avoid=avoid))
            i += 1
        return best_angles

    def optimize_angle(self, voi, couch, gantry, margin, iteration, avoid=[]):
        """ TODO: Documentation
        """
        if iteration is 0:
            return [gantry, couch]
        grid = self.calculate_quality_list(
            voi, np.linspace(gantry, gantry + margin, 3), np.linspace(couch, couch + margin, 3), avoid=avoid)
        min_item = min(grid, key=lambda x: x["data"][0])
        return self.optimize_angle(voi, min_item["gantry"], min_item["couch"], margin / 2, iteration - 1, avoid=avoid)

    def calculate_angle_quality_thread(self,
                                       voi,
                                       gantry,
                                       couch,
                                       calculate_from=0,
                                       stepsize=1.0,
                                       q=None,
                                       avoid=[],
                                       voi_cube=None,
                                       gradient=True):
        """ TODO: Documentation
        """
        os.nice(1)
        for couch_angle in couch:
            qual = self.calculate_angle_quality(voi, gantry, couch_angle, calculate_from, stepsize, avoid, voi_cube,
                                                gradient)
            q.put({"couch": couch_angle, "gantry": gantry, "data": qual})

    def calculate_angle_quality(self,
                                voi,
                                gantry,
                                couch,
                                calculate_from=0,
                                stepsize=1.0,
                                avoid=[],
                                voi_cube=None,
                                gradient=True):
        """
        Calculates a quality index for a given gantry/couch combination.
        """
        voi_min, voi_max = voi.get_min_max()
        for v in avoid:
            v_min, v_max = v.get_min_max()
        if voi_cube is None:
            d = voi.get_voi_cube()
            d.cube = np.array(d.cube, dtype=np.float32)
            voi_cube = DensityProjections(d)

        data, start, basis = self.calculate_projection(voi, gantry, couch, calculate_from, stepsize)
        voi_proj, t1, t2 = voi_cube.calculate_projection(voi, gantry, couch, 1, stepsize)

        if gradient:
            gradient = np.gradient(data)
            data = (gradient[0]**2 + gradient[1]**2)**0.5
        a = data * (voi_proj > 0.0)
        quality = sum(a)
        area = sum(voi_proj > 0.0)
        # ~ area = sum(data>0.0)/10
        if gradient:
            mean_quality = 10 - abs(quality / area)
        else:
            mean_quality = abs(quality / area)
        return mean_quality, quality, area

    def calculate_projection(self, voi, gantry, couch, calculate_from=0, stepsize=1.0):
        """
        TODO: documentation

        :param Voi voi: tumortarget, type is Voi
        :param float gantry: angle in degrees
        :param float couch: angle in degrees
        :param int calculate_from: 0 is mass center 1 is the most distant point in the tumor from the beamaxis
        :param float stepsize: relative to pixelsize, 1 is a step of 1 pixel
        :return:
        """
        # min_structure = 5 TODO why not used ?
        basis = get_basis_from_angles(gantry, couch)
        # Convert angles from degrees to radians
        gantry /= 180.0 / np.pi
        couch /= 180.0 / np.pi
        # calculate surface normal
        step_vec = -stepsize * self.cube.pixel_size * np.array(basis[0])
        step_length = stepsize * self.cube.pixel_size
        center = voi.calculate_center()
        # ~ (b,c) = self.calculate_plane_vectors(gantry,couch)
        b = basis[1]
        c = basis[2]
        min_window, max_window = voi.get_min_max()
        size = np.array(max_window) - np.array(min_window)
        window_size = np.array([((sin(couch) * size[1])**2 + (cos(couch) * size[2])**2)**0.5, (
            (sin(gantry) * size[0])**2 + (cos(gantry) * size[1])**2 + (sin(couch) * size[2])**2)**0.5]) * 2

        dimension = window_size / self.cube.pixel_size
        dimension = np.int16(dimension)
        start = center - self.cube.pixel_size * 0.5 * np.array(dimension[0] * b + dimension[1] * c)
        if calculate_from is 1:
            start = self.calculate_back_start_voi(voi, start, step_vec)
        if calculate_from is 2:
            start = self.calculate_front_start_voi(voi, start, step_vec)

        data = pytriplib.calculate_wepl(
            self.cube.cube, np.array(start), np.array(basis) * step_length, dimension,
            np.array([self.cube.pixel_size, self.cube.pixel_size, self.cube.slice_distance]))
        data *= step_length
        return data, start, [b, c]

    def calculate_back_start_voi(self, voi, start, beam_axis):
        """ TODO: Documentation

        :params voi:
        :params start:
        :params beam axis:
        :returns:
        """
        points = voi.get_3d_polygon() - start
        distance = min(np.dot(points, beam_axis))
        return start + distance * beam_axis

    def calculate_front_start_voi(self, voi, start, beam_axis):
        """ TODO: Documentation

        :params voi:
        :params start:
        :params beam axis:
        :returns:
        """
        points = voi.get_3d_polygon() - start
        distance = max(np.dot(points, beam_axis))
        return start + distance * beam_axis


class DensityCube(Cube):
    """ Class for working with denisity cubes [g/cm3]
    """

    def __init__(self, ctxcube, hlut_path="/data/hlut_den.dat"):
        """ Creates a DensityCube based on a CTX cube and a Hounsfield lookup table.

        :params CtxCube ctxcube:
        :params str hlut_path: path to Hounsfield lookup table, relative from where pytrip was executed.
        """
        self.ctxcube = ctxcube
        super(DensityCube, self).__init__(ctxcube)
        self.type = "Density"
        self.directory = os.path.dirname(os.path.abspath(__file__))
        self.hlut_file = os.path.join(self.directory, hlut_path)
        self.import_hlut()
        self.calculate_cube()
        self.ctxcube = None

    def calculate_cube(self):
        """
        Calculate the density values from HU table and interpolating the loaded hlut_data.
        """
        ctxdata = self.ctxcube.cube
        ctxdata = ctxdata.reshape(self.dimx * self.dimy * self.dimz)
        gc.collect()

        cube = interpolate.splev(ctxdata, self.hlut_data)
        # ~ cube = self.hlut_data(ctxdata)
        cube = cube.astype(np.float32)
        self.cube = np.reshape(cube, (self.dimz, self.dimy, self.dimx))

    def import_hlut(self):
        """ Imports the Hounsfield lookup table and stores it into self.hlut_data
        """
        fp = open(self.hlut_file, "r")
        lines = fp.read()
        fp.close()
        lines = lines.split('\n')
        x_data = []
        y_data = []
        for line in lines:
            a = line.split()
            if len(a):
                x_data.append(float(a[0]))
                y_data.append(float(a[3]))
        self.hlut_data = interpolate.splrep(np.array(x_data), np.array(y_data), k=1)



#
#    Copyright (C) 2010-2016 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Simple example of how to do arithmetic on Cube objects in PyTRiP.
"""
import pytrip as pt

# sum two dose cubes, write result:
print("Two half boxes: out.dos")
d1 = pt.DosCube()
d2 = pt.DosCube()
d1.read("box052000.dos")
d2.read("box053000.dos")
d = (d1 + d2)
d.write("out.dos")

# print minium and maximum value found in cubes
print(d1.cube.min(), d1.cube.max())
print(d2.cube.min(), d2.cube.max())

# calculate new dose average LET cube
l1 = pt.LETCube()
l2 = pt.LETCube()
l1.read("box052000.dosemlet.dos")
l2.read("box053000.dosemlet.dos")

l = ((d1 * l1) + (d2 * l2)) / (d1 + d2)
l.write("out.dosemlet.dos")
